# Copyright 2022 Lenovo Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pyghmi.redfish.oem.generic as generic


class OEMHandler(generic.OEMHandler):

    def __init__(self, sysinfo, sysurl, webclient, cache, gpool=None):
        super(OEMHandler, self).__init__(sysinfo, sysurl, webclient, cache,
                                         gpool)

    def set_bootdev(self, bootdev, persist=False, uefiboot=None,
                    fishclient=None):
        # gleaned from web console, under configuration, system settings,
        # hardware, first boot device. iDrac presumes that the standard
        # explicitly refers only to physical devices. I think the intent
        # is the exact opposite for 'removable' media, and thus redirect
        # the 'physical' standard to the vFDD/VCD-DVD seen in the idrac
        # web gui
        if bootdev not in ('floppy', 'cd'):
            return super(OEMHandler, self).set_bootdev(bootdev, persist,
                                                       uefiboot, fishclient)
        payload = {'Attributes': {}}
        if persist:
            payload['Attributes']['ServerBoot.1.BootOnce'] = 'Disabled'
        else:
            payload['Attributes']['ServerBoot.1.BootOnce'] = 'Enabled'
        if bootdev == 'floppy':
            payload['Attributes']['ServerBoot.1.FirstBootDevice'] = 'vFDD'
        elif bootdev == 'cd':
            payload['Attributes']['ServerBoot.1.FirstBootDevice'] = 'VCD-DVD'
        fishclient._do_web_request(
            '/redfish/v1/Managers/iDRAC.Embedded.1/Attributes',
            payload, method='PATCH')
        return {'bootdev': bootdev}



# Copyright 2015 Lenovo
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This implements parsing of DDR SPD data.  This is offered up in a pass
through fashion by some service processors.

For now, just doing DDR3 and DDR4

In many cases, astute readers will note that some of the lookup tables
should be a matter of math rather than lookup.  However the SPD
specification explicitly reserves values not in the lookup tables for
future use.  It has happened, for example, that a spec was amended
with discontinuous values for a field that was until that point
possible to derive in a formulaic way
"""

import math
import struct

jedec_ids = [
    {
        0x01: "AMD",
        0x02: "AMI",
        0x83: "Fairchild",
        0x04: "Fujitsu",
        0x85: "GTE",
        0x86: "Harris",
        0x07: "Hitachi",
        0x08: "Inmos",
        0x89: "Intel",
        0x8a: "I.T.T.",
        0x0b: "Intersil",
        0x8c: "Monolithic Memories",
        0x0d: "Mostek",
        0x0e: "Motorola",
        0x8f: "National",
        0x10: "NEC",
        0x91: "RCA",
        0x92: "Raytheon",
        0x13: "Conexant (Rockwell)",
        0x94: "Seeq",
        0x15: "Philips Semi. (Signetics)",
        0x16: "Synertek",
        0x97: "Texas Instruments",
        0x98: "Toshiba",
        0x19: "Xicor",
        0x1a: "Zilog",
        0x9b: "Eurotechnique",
        0x1c: "Mitsubishi",
        0x9d: "Lucent (AT&T)",
        0x9e: "Exel",
        0x1f: "Atmel",
        0x20: "SGS/Thomson",
        0xa1: "Lattice Semi.",
        0xa2: "NCR",
        0x23: "Wafer Scale Integration",
        0xa4: "IBM",
        0x25: "Tristar",
        0x26: "Visic",
        0xa7: "Intl. CMOS Technology",
        0xa8: "SSSI",
        0x29: "Microchip Technology",
        0x2a: "Ricoh Ltd.",
        0xab: "VLSI",
        0x2c: "Micron Technology",
        0xad: "Hyundai Electronics",
        0xae: "OKI Semiconductor",
        0x2f: "ACTEL",
        0xb0: "Sharp",
        0x31: "Catalyst",
        0x32: "Panasonic",
        0xb3: "IDT",
        0x34: "Cypress",
        0xb5: "DEC",
        0xb6: "LSI Logic",
        0x37: "Zarlink",
        0x38: "UTMC",
        0xb9: "Thinking Machine",
        0xba: "Thomson CSF",
        0x3b: "Integrated CMOS(Vertex)",
        0xbc: "Honeywell",
        0x3d: "Tektronix",
        0x3e: "Sun Microsystems",
        0xbf: "SST",
        0x40: "MOSEL",
        0xc1: "Infineon",
        0xc2: "Macronix",
        0x43: "Xerox",
        0xc4: "Plus Logic",
        0x45: "SunDisk",
        0x46: "Elan Circuit Tech.",
        0xc7: "European Silicon Str.",
        0xc8: "Apple Computer",
        0xc9: "Xilinx",
        0x4a: "Compaq",
        0xcb: "Protocol Engines",
        0x4c: "SCI",
        0xcd: "Seiko Instruments",
        0xce: "Samsung",
        0x4f: "I3 Design System",
        0xd0: "Klic",
        0x51: "Crosspoint Solutions",
        0x52: "Alliance Semiconductor",
        0xd3: "Tandem",
        0x54: "Hewlett-Packard",
        0xd5: "Intg. Silicon Solutions",
        0xd6: "Brooktree",
        0x57: "New Media",
        0x58: "MHS Electronic",
        0xd9: "Performance Semi.",
        0xda: "Winbond Electronic",
        0x5b: "Kawasaki Steel",
        0xdc: "Bright Micro",
        0x5d: "TECMAR",
        0x5e: "Exar",
        0xdf: "PCMCIA",
        0xe0: "LG Semiconductor",
        0x61: "Northern Telecom",
        0x62: "Sanyo",
        0xe3: "Array Microsystems",
        0x64: "Crystal Semiconductor",
        0xe5: "Analog Devices",
        0xe6: "PMC-Sierra",
        0x67: "Asparix",
        0x68: "Convex Computer",
        0xe9: "Quality Semiconductor",
        0xea: "Nimbus Technology",
        0x6b: "Transwitch",
        0xec: "Micronas (ITT Intermetall)",
        0x6d: "Cannon",
        0x6e: "Altera",
        0xef: "NEXCOM",
        0x70: "QUALCOMM",
        0xf1: "Sony",
        0xf2: "Cray Research",
        0x73: "AMS (Austria Micro)",
        0xf4: "Vitesse",
        0x75: "Aster Electronics",
        0x76: "Bay Networks (Synoptic)",
        0xf7: "Zentrum",
        0xf8: "TRW",
        0x79: "Thesys",
        0x7a: "Solbourne Computer",
        0xfb: "Allied-Signal",
        0x7c: "Dialog",
        0xfd: "Media Vision",
        0xfe: "Level One Communication",
    },
    {
        0x01: "Cirrus Logic",
        0x02: "National Instruments",
        0x83: "ILC Data Device",
        0x04: "Alcatel Mietec",
        0x85: "Micro Linear",
        0x86: "Univ. of NC",
        0x07: "JTAG Technologies",
        0x08: "Loral",
        0x89: "Nchip",
        0x8A: "Galileo Tech",
        0x0B: "Bestlink Systems",
        0x8C: "Graychip",
        0x0D: "GENNUM",
        0x0E: "VideoLogic",
        0x8F: "Robert Bosch",
        0x10: "Chip Express",
        0x91: "DATARAM",
        0x92: "United Microelec Corp.",
        0x13: "TCSI",
        0x94: "Smart Modular",
        0x15: "Hughes Aircraft",
        0x16: "Lanstar Semiconductor",
        0x97: "Qlogic",
        0x98: "Kingston",
        0x19: "Music Semi",
        0x1A: "Ericsson Components",
        0x9B: "SpaSE",
        0x1C: "Eon Silicon Devices",
        0x9D: "Programmable Micro Corp",
        0x9E: "DoD",
        0x1F: "Integ. Memories Tech.",
        0x20: "Corollary Inc.",
        0xA1: "Dallas Semiconductor",
        0xA2: "Omnivision",
        0x23: "EIV(Switzerland)",
        0xA4: "Novatel Wireless",
        0x25: "Zarlink (formerly Mitel)",
        0x26: "Clearpoint",
        0xA7: "Cabletron",
        0xA8: "Silicon Technology",
        0x29: "Vanguard",
        0x2A: "Hagiwara Sys-Com",
        0xAB: "Vantis",
        0x2C: "Celestica",
        0xAD: "Century",
        0xAE: "Hal Computers",
        0x2F: "Rohm Company Ltd.",
        0xB0: "Juniper Networks",
        0x31: "Libit Signal Processing",
        0x32: "Enhanced Memories Inc.",
        0xB3: "Tundra Semiconductor",
        0x34: "Adaptec Inc.",
        0xB5: "LightSpeed Semi.",
        0xB6: "ZSP Corp.",
        0x37: "AMIC Technology",
        0x38: "Adobe Systems",
        0xB9: "Dynachip",
        0xBA: "PNY Electronics",
        0x3B: "Newport Digital",
        0xBC: "MMC Networks",
        0x3D: "T Square",
        0x3E: "Seiko Epson",
        0xBF: "Broadcom",
        0x40: "Viking Components",
        0xC1: "V3 Semiconductor",
        0xC2: "Flextronics (formerly Orbit)",
        0x43: "Suwa Electronics",
        0xC4: "Transmeta",
        0x45: "Micron CMS",
        0x46: "American Computer & Digital Components Inc",
        0xC7: "Enhance 3000 Inc",
        0xC8: "Tower Semiconductor",
        0x49: "CPU Design",
        0x4A: "Price Point",
        0xCB: "Maxim Integrated Product",
        0x4C: "Tellabs",
        0xCD: "Centaur Technology",
        0xCE: "Unigen Corporation",
        0x4F: "Transcend Information",
        0xD0: "Memory Card Technology",
        0x51: "CKD Corporation Ltd.",
        0x52: "Capital Instruments, Inc.",
        0xD3: "Aica Kogyo, Ltd.",
        0x54: "Linvex Technology",
        0xD5: "MSC Vertriebs GmbH",
        0xD6: "AKM Company, Ltd.",
        0x57: "Dynamem, Inc.",
        0x58: "NERA ASA",
        0xD9: "GSI Technology",
        0xDA: "Dane-Elec (C Memory)",
        0x5B: "Acorn Computers",
        0xDC: "Lara Technology",
        0x5D: "Oak Technology, Inc.",
        0x5E: "Itec Memory",
        0xDF: "Tanisys Technology",
        0xE0: "Truevision",
        0x61: "Wintec Industries",
        0x62: "Super PC Memory",
        0xE3: "MGV Memory",
        0x64: "Galvantech",
        0xE5: "Gadzoox Nteworks",
        0xE6: "Multi Dimensional Cons.",
        0x67: "GateField",
        0x68: "Integrated Memory System",
        0xE9: "Triscend",
        0xEA: "XaQti",
        0x6B: "Goldenram",
        0xEC: "Clear Logic",
        0x6D: "Cimaron Communications",
        0x6E: "Nippon Steel Semi. Corp.",
        0xEF: "Advantage Memory",
        0x70: "AMCC",
        0xF1: "LeCroy",
        0xF2: "Yamaha Corporation",
        0x73: "Digital Microwave",
        0xF4: "NetLogic Microsystems",
        0x75: "MIMOS Semiconductor",
        0x76: "Advanced Fibre",
        0xF7: "BF Goodrich Data.",
        0xF8: "Epigram",
        0x79: "Acbel Polytech Inc.",
        0x7A: "Apacer Technology",
        0xFB: "Admor Memory",
        0x7C: "FOXCONN",
        0xFD: "Quadratics Superconductor",
        0xFE: "3COM",
    },
    {
        0x01: "Camintonn Corporation",
        0x02: "ISOA Incorporated",
        0x83: "Agate Semiconductor",
        0x04: "ADMtek Incorporated",
        0x85: "HYPERTEC",
        0x86: "Adhoc Technologies",
        0x07: "MOSAID Technologies",
        0x08: "Ardent Technologies",
        0x89: "Switchcore",
        0x8A: "Cisco Systems, Inc.",
        0x0B: "Allayer Technologies",
        0x8C: "WorkX AG",
        0x0D: "Oasis Semiconductor",
        0x0E: "Novanet Semiconductor",
        0x8F: "E-M Solutions",
        0x10: "Power General",
        0x91: "Advanced Hardware Arch.",
        0x92: "Inova Semiconductors GmbH",
        0x13: "Telocity",
        0x94: "Delkin Devices",
        0x15: "Symagery Microsystems",
        0x16: "C-Port Corporation",
        0x97: "SiberCore Technologies",
        0x98: "Southland Microsystems",
        0x19: "Malleable Technologies",
        0x1A: "Kendin Communications",
        0x9B: "Great Technology Microcomputer",
        0x1C: "Sanmina Corporation",
        0x9D: "HADCO Corporation",
        0x9E: "Corsair",
        0x1F: "Actrans System Inc.",
        0x20: "ALPHA Technologies",
        0xA1: "Cygnal Integrated Products Incorporated",
        0xA2: "Artesyn Technologies",
        0x23: "Align Manufacturing",
        0xA4: "Peregrine Semiconductor",
        0x25: "Chameleon Systems",
        0x26: "Aplus Flash Technology",
        0xA7: "MIPS Technologies",
        0xA8: "Chrysalis ITS",
        0x29: "ADTEC Corporation",
        0x2A: "Kentron Technologies",
        0xAB: "Win Technologies",
        0x2C: "ASIC Designs Inc",
        0xAD: "Extreme Packet Devices",
        0xAE: "RF Micro Devices",
        0x2F: "Siemens AG",
        0xB0: "Sarnoff Corporation",
        0x31: "Itautec Philco SA",
        0x32: "Radiata Inc.",
        0xB3: "Benchmark Elect. (AVEX)",
        0x34: "Legend",
        0xB5: "SpecTek Incorporated",
        0xB6: "Hi/fn",
        0x37: "Enikia Incorporated",
        0x38: "SwitchOn Networks",
        0xB9: "AANetcom Incorporated",
        0xBA: "Micro Memory Bank",
        0x3B: "ESS Technology",
        0xBC: "Virata Corporation",
        0x3D: "Excess Bandwidth",
        0x3E: "West Bay Semiconductor",
        0xBF: "DSP Group",
        0x40: "Newport Communications",
        0xC1: "Chip2Chip Incorporated",
        0xC2: "Phobos Corporation",
        0x43: "Intellitech Corporation",
        0xC4: "Nordic VLSI ASA",
        0x45: "Ishoni Networks",
        0x46: "Silicon Spice",
        0xC7: "Alchemy Semiconductor",
        0xC8: "Agilent Technologies",
        0x49: "Centillium Communications",
        0x4A: "W.L. Gore",
        0xCB: "HanBit Electronics",
        0x4C: "GlobeSpan",
        0xCD: "Element 14",
        0xCE: "Pycon",
        0x4F: "Saifun Semiconductors",
        0xD0: "Sibyte, Incorporated",
        0x51: "MetaLink Technologies",
        0x52: "Feiya Technology",
        0xD3: "I & C Technology",
        0x54: "Shikatronics",
        0xD5: "Elektrobit",
        0xD6: "Megic",
        0x57: "Com-Tier",
        0x58: "Malaysia Micro Solutions",
        0xD9: "Hyperchip",
        0xDA: "Gemstone Communications",
        0x5B: "Anadyne Microelectronics",
        0xDC: "3ParData",
        0x5D: "Mellanox Technologies",
        0x5E: "Tenx Technologies",
        0xDF: "Helix AG",
        0xE0: "Domosys",
        0x61: "Skyup Technology",
        0x62: "HiNT Corporation",
        0xE3: "Chiaro",
        0x64: "MCI Computer GMBH",
        0xE5: "Exbit Technology A/S",
        0xE6: "Integrated Technology Express",
        0x67: "AVED Memory",
        0x68: "Legerity",
        0xE9: "Jasmine Networks",
        0xEA: "Caspian Networks",
        0x6B: "nCUBE",
        0xEC: "Silicon Access Networks",
        0x6D: "FDK Corporation",
        0x6E: "High Bandwidth Access",
        0xEF: "MultiLink Technology",
        0x70: "BRECIS",
        0xF1: "World Wide Packets",
        0xF2: "APW",
        0x73: "Chicory Systems",
        0xF4: "Xstream Logic",
        0x75: "Fast-Chip",
        0x76: "Zucotto Wireless",
        0xF7: "Realchip",
        0xF8: "Galaxy Power",
        0x79: "eSilicon",
        0x7A: "Morphics Technology",
        0xFB: "Accelerant Networks",
        0x7C: "Silicon Wave",
        0xFD: "SandCraft",
        0xFE: "Elpida",
    },
    {
        0x01: "Solectron",
        0x02: "Optosys Technologies",
        0x83: "Buffalo (Formerly Melco)",
        0x04: "TriMedia Technologies",
        0x85: "Cyan Technologies",
        0x86: "Global Locate",
        0x07: "Optillion",
        0x08: "Terago Communications",
        0x89: "Ikanos Communications",
        0x8A: "Princeton Technology",
        0x0B: "Nanya Technology",
        0x8C: "Elite Flash Storage",
        0x0D: "Mysticom",
        0x0E: "LightSand Communications",
        0x8F: "ATI Technologies",
        0x10: "Agere Systems",
        0x91: "NeoMagic",
        0x92: "AuroraNetics",
        0x13: "Golden Empire",
        0x94: "Muskin",
        0x15: "Tioga Technologies",
        0x16: "Netlist",
        0x97: "TeraLogic",
        0x98: "Cicada Semiconductor",
        0x19: "Centon Electronics",
        0x1A: "Tyco Electronics",
        0x9B: "Magis Works",
        0x1C: "Zettacom",
        0x9D: "Cogency Semiconductor",
        0x9E: "Chipcon AS",
        0x1F: "Aspex Technology",
        0x20: "F5 Networks",
        0xA1: "Programmable Silicon Solutions",
        0xA2: "ChipWrights",
        0x23: "Acorn Networks",
        0xA4: "Quicklogic",
        0x25: "Kingmax Semiconductor",
        0x26: "BOPS",
        0xA7: "Flasys",
        0xA8: "BitBlitz Communications",
        0x29: "eMemory Technology",
        0x2A: "Procket Networks",
        0xAB: "Purple Ray",
        0x2C: "Trebia Networks",
        0xAD: "Delta Electronics",
        0xAE: "Onex Communications",
        0x2F: "Ample Communications",
        0xB0: "Memory Experts Intl",
        0x31: "Astute Networks",
        0x32: "Azanda Network Devices",
        0xB3: "Dibcom",
        0x34: "Tekmos",
        0xB5: "API NetWorks",
        0xB6: "Bay Microsystems",
        0x37: "Firecron Ltd",
        0x38: "Resonext Communications",
        0xB9: "Tachys Technologies",
        0xBA: "Equator Technology",
        0x3B: "Concept Computer",
        0xBC: "SILCOM",
        0x3D: "3Dlabs",
        0x3E: "ct Magazine",
        0xBF: "Sanera Systems",
        0x40: "Silicon Packets",
        0xC1: "Viasystems Group",
        0xC2: "Simtek",
        0x43: "Semicon Devices Singapore",
        0xC4: "Satron Handelsges",
        0x45: "Improv Systems",
        0x46: "INDUSYS GmbH",
        0xC7: "Corrent",
        0xC8: "Infrant Technologies",
        0x49: "Ritek Corp",
        0x4A: "empowerTel Networks",
        0xCB: "Hypertec",
        0x4C: "Cavium Networks",
        0xCD: "PLX Technology",
        0xCE: "Massana Design",
        0x4F: "Intrinsity",
        0xD0: "Valence Semiconductor",
        0x51: "Terawave Communications",
        0x52: "IceFyre Semiconductor",
        0xD3: "Primarion",
        0x54: "Picochip Designs Ltd",
        0xD5: "Silverback Systems",
        0xD6: "Jade Star Technologies",
        0x57: "Pijnenburg Securealink",
        0x58: "MemorySolutioN",
        0xD9: "Cambridge Silicon Radio",
        0xDA: "Swissbit",
        0x5B: "Nazomi Communications",
        0xDC: "eWave System",
        0x5D: "Rockwell Collins",
        0x5E: "PAION",
        0xDF: "Alphamosaic Ltd",
        0xE0: "Sandburst",
        0x61: "SiCon Video",
        0x62: "NanoAmp Solutions",
        0xE3: "Ericsson Technology",
        0x64: "PrairieComm",
        0xE5: "Mitac International",
        0xE6: "Layer N Networks",
        0x67: "Atsana Semiconductor",
        0x68: "Allegro Networks",
        0xE9: "Marvell Semiconductors",
        0xEA: "Netergy Microelectronic",
        0x6B: "NVIDIA",
        0xEC: "Internet Machines",
        0x6D: "Peak Electronics",
        0xEF: "Accton Technology",
        0x70: "Teradiant Networks",
        0xF1: "Europe Technologies",
        0xF2: "Cortina Systems",
        0x73: "RAM Components",
        0xF4: "Raqia Networks",
        0x75: "ClearSpeed",
        0x76: "Matsushita Battery",
        0xF7: "Xelerated",
        0xF8: "SimpleTech",
        0x79: "Utron Technology",
        0x7A: "Astec International",
        0xFB: "AVM gmbH",
        0x7C: "Redux Communications",
        0xFD: "Dot Hill Systems",
        0xFE: "TeraChip",
    },
    {
        0x01: "T-RAM Incorporated",
        0x02: "Innovics Wireless",
        0x83: "Teknovus",
        0x04: "KeyEye Communications",
        0x85: "Runcom Technologies",
        0x86: "RedSwitch",
        0x07: "Dotcast",
        0x08: "Silicon Mountain Memory",
        0x89: "Signia Technologies",
        0x8A: "Pixim",
        0x0B: "Galazar Networks",
        0x8C: "White Electronic Designs",
        0x0D: "Patriot Scientific",
        0x0E: "Neoaxiom Corporation",
        0x8F: "3Y Power Technology",
        0x10: "Europe Technologies",
        0x91: "Potentia Power Systems",
        0x92: "C-guys Incorporated",
        0x13: "Digital Communications Technology Incorporated",
        0x94: "Silicon-Based Technology",
        0x15: "Fulcrum Microsystems",
        0x16: "Positivo Informatica Ltd",
        0x97: "XIOtech Corporation",
        0x98: "PortalPlayer",
        0x19: "Zhiying Software",
        0x1A: "Direct2Data",
        0x9B: "Phonex Broadband",
        0x1C: "Skyworks Solutions",
        0x9D: "Entropic Communications",
        0x9E: "Pacific Force Technology",
        0x1F: "Zensys A/S",
        0x20: "Legend Silicon Corp.",
        0xA1: "sci-worx GmbH",
        0xA2: "Oasis Silicon Systems",
        0x23: "Renesas Technology",
        0xA4: "Raza Microelectronics",
        0x25: "Phyworks",
        0x26: "MediaTek",
        0xA7: "Non-cents Productions",
        0xA8: "US Modular",
        0x29: "Wintegra Ltd",
        0x2A: "Mathstar",
        0xAB: "StarCore",
        0x2C: "Oplus Technologies",
        0xAD: "Mindspeed",
        0xAE: "Just Young Computer",
        0x2F: "Radia Communications",
        0xB0: "OCZ",
        0x31: "Emuzed",
        0x32: "LOGIC Devices",
        0xB3: "Inphi Corporation",
        0x34: "Quake Technologies",
        0xB5: "Vixel",
        0xB6: "SolusTek",
        0x37: "Kongsberg Maritime",
        0x38: "Faraday Technology",
        0xB9: "Altium Ltd.",
        0xBA: "Insyte",
        0x3B: "ARM Ltd.",
        0xBC: "DigiVision",
        0x3D: "Vativ Technologies",
        0x3E: "Endicott Interconnect Technologies",
        0xBF: "Pericom",
        0x40: "Bandspeed",
        0xC1: "LeWiz Communications",
        0xC2: "CPU Technology",
        0x43: "Ramaxel Technology",
        0xC4: "DSP Group",
        0x45: "Axis Communications",
        0x46: "Legacy Electronics",
        0xC7: "Chrontel",
        0xC8: "Powerchip Semiconductor",
        0x49: "MobilEye Technologies",
        0x4A: "Excel Semiconductor",
        0xCB: "A-DATA Technology",
        0x4C: "VirtualDigm",
    },
]

memory_types = {
    1: "STD FPM DRAM",
    2: "EDO",
    3: "Pipelined Nibble",
    4: "SDRAM",
    5: "ROM",
    6: "DDR SGRAM",
    7: "DDR SDRAM",
    8: "DDR2 SDRAM",
    9: "DDR2 SDRAM FB-DIMM",
    10: "DDR2 SDRAM FB-DIMM PROBE",
    11: "DDR3 SDRAM",
    12: "DDR4 SDRAM",
}

module_types = {
    1: "RDIMM",
    2: "UDIMM",
    3: "SODIMM",
    4: "Micro-DIMM",
    5: "Mini-RDIMM",
    6: "Mini-UDIMM",
}

ddr3_module_capacity = {
    0: 256,
    1: 512,
    2: 1024,
    3: 2048,
    4: 4096,
    5: 8192,
    6: 16384,
    7: 32768,
}

ddr3_dev_width = {
    0: 4,
    1: 8,
    2: 16,
    3: 32,
}

ddr3_ranks = {
    0: 1,
    1: 2,
    2: 3,
    3: 4
}

ddr3_bus_width = {
    0: 8,
    1: 16,
    2: 32,
    3: 64,
}


def speed_from_clock(clock):
    return int(clock * 8 - (clock * 8 % 100))


def decode_manufacturer(index, mfg):
    index &= 0x7f
    try:
        return jedec_ids[index][mfg]
    except (KeyError, IndexError):
        return 'Unknown ({0}, {1})'.format(index, mfg)


def decode_spd_date(year, week):
    if year == 0 and week == 0:
        return 'Unknown'
    return '20{0:02x}-W{1:x}'.format(year, week)


class SPD(object):
    def __init__(self, bytedata):
        """Parsed memory information

        Parse bytedata input and provide a structured detail about the
        described memory component

        :param bytedata: A bytearray of data to decode
        :return:
        """
        self.rawdata = bytearray(bytedata)
        spd = self.rawdata
        self.info = {'memory_type': memory_types.get(spd[2], 'Unknown')}
        if spd[2] == 11:
            self._decode_ddr3()
        elif spd[2] == 12:
            self._decode_ddr4()

    def _decode_ddr3(self):
        spd = self.rawdata
        finetime = (spd[9] >> 4) / (spd[9] & 0xf)
        fineoffset = spd[34]
        if fineoffset & 0b10000000:
            # Take two's complement for negative offset
            fineoffset = 0 - ((fineoffset ^ 0xff) + 1)
        fineoffset = (finetime * fineoffset) * 10 ** -3
        mtb = spd[10] / float(spd[11])
        clock = math.floor(2 / ((mtb * spd[12] + fineoffset) * 10 ** -3))
        self.info['speed'] = speed_from_clock(clock)
        self.info['ecc'] = (spd[8] & 0b11000) != 0
        self.info['module_type'] = module_types.get(spd[3] & 0xf, 'Unknown')
        sdramcap = ddr3_module_capacity[spd[4] & 0xf]
        buswidth = ddr3_bus_width[spd[8] & 0b111]
        sdramwidth = ddr3_dev_width[spd[7] & 0b111]
        ranks = ddr3_ranks[(spd[7] & 0b111000) >> 3]
        self.info['capacity_mb'] = sdramcap / 8 * buswidth / sdramwidth * ranks
        self.info['manufacturer'] = decode_manufacturer(spd[117], spd[118])
        self.info['manufacture_location'] = spd[119]
        self.info['manufacture_date'] = decode_spd_date(spd[120], spd[121])
        self.info['serial'] = hex(struct.unpack(
            '>I', struct.pack('4B', *spd[122:126]))[0])[2:].rjust(8, '0')
        self.info['model'] = struct.pack('20B', *spd[128:148]).strip(
            b'\x00\xff ')

    def _decode_ddr4(self):
        spd = self.rawdata
        if spd[17] == 0:
            fineoffset = spd[125]
            if fineoffset & 0b10000000:
                fineoffset = 0 - ((fineoffset ^ 0xff) + 1)
            clock = math.floor(
                2 / ((0.125 * spd[18] + fineoffset * 0.001) * 0.001))
            self.info['speed'] = speed_from_clock(clock)
        else:
            self.info['speed'] = 'Unknown'
        self.info['ecc'] = (spd[13] & 0b11000) == 0b1000
        self.info['module_type'] = module_types.get(spd[3] & 0xf,
                                                    'Unknown')
        sdramcap = ddr3_module_capacity[spd[4] & 0xf]
        buswidth = ddr3_bus_width[spd[13] & 0b111]
        sdramwidth = ddr3_dev_width[spd[12] & 0b111]
        ranks = ddr3_ranks[(spd[12] & 0b111000) >> 3]
        if spd[6] & 0b11 == 0b10:
            ranks = ranks * (((spd[6] >> 4) & 0b111) + 1)
        self.info['capacity_mb'] = sdramcap / 8 * buswidth / sdramwidth * ranks
        self.info['manufacturer'] = decode_manufacturer(spd[320], spd[321])
        self.info['manufacture_location'] = spd[322]
        self.info['manufacture_date'] = decode_spd_date(spd[323], spd[324])
        self.info['serial'] = hex(struct.unpack(
            '>I', struct.pack('4B', *spd[325:329]))[0])[2:].rjust(8, '0')
        self.info['model'] = struct.pack('20B', *spd[329:349]).strip(
            b'\x00\xff ')



# Copyright 2015-2017 Lenovo
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import binascii
import socket
import struct
import traceback
import weakref

import pyghmi.constants as pygconst
import pyghmi.exceptions as pygexc
import pyghmi.ipmi.oem.generic as generic
from pyghmi.ipmi.oem.lenovo import cpu
from pyghmi.ipmi.oem.lenovo import dimm
from pyghmi.ipmi.oem.lenovo import drive
from pyghmi.ipmi.oem.lenovo import energy
from pyghmi.ipmi.oem.lenovo import firmware
from pyghmi.ipmi.oem.lenovo import imm
from pyghmi.ipmi.oem.lenovo import inventory
from pyghmi.ipmi.oem.lenovo import nextscale
from pyghmi.ipmi.oem.lenovo import pci
from pyghmi.ipmi.oem.lenovo import psu
from pyghmi.ipmi.oem.lenovo import raid_controller
from pyghmi.ipmi.oem.lenovo import raid_drive
import pyghmi.ipmi.private.constants as ipmiconst
import pyghmi.ipmi.private.util as util
import pyghmi.redfish.command as redfishcmd
from pyghmi.redfish.oem.lenovo import tsma
import pyghmi.util.webclient as wc

try:
    from urllib import urlencode
except ImportError:
    from urllib.parse import urlencode

try:
    range = xrange
except NameError:
    pass
try:
    buffer
except NameError:
    buffer = memoryview

inventory.register_inventory_category(cpu)
inventory.register_inventory_category(dimm)
inventory.register_inventory_category(pci)
inventory.register_inventory_category(drive)
inventory.register_inventory_category(psu)
inventory.register_inventory_category(raid_drive)
inventory.register_inventory_category(raid_controller)


firmware_types = {
    1: 'Management Controller',
    2: 'UEFI/BIOS',
    3: 'CPLD',
    4: 'Power Supply',
    5: 'Storage Adapter',
    6: 'Add-in Adapter',
}

firmware_event = {
    0: ('Update failed', pygconst.Health.Failed),
    1: ('Update succeeded', pygconst.Health.Ok),
    2: ('Update aborted', pygconst.Health.Ok),
    3: ('Unknown', pygconst.Health.Warning),
}

me_status = {
    0: ('Recovery GPIO forced', pygconst.Health.Warning),
    1: ('ME Image corrupt', pygconst.Health.Critical),
    2: ('Flash erase error', pygconst.Health.Critical),
    3: ('Unspecified flash state', pygconst.Health.Warning),
    4: ('ME watchdog timeout', pygconst.Health.Critical),
    5: ('ME platform reboot', pygconst.Health.Critical),
    6: ('ME update', pygconst.Health.Ok),
    7: ('Manufacturing error', pygconst.Health.Critical),
    8: ('ME Flash storage integrity error', pygconst.Health.Critical),
    9: ('ME firmware exception', pygconst.Health.Critical),  # event data 3..
    0xa: ('ME firmware worn', pygconst.Health.Warning),
    0xc: ('Invalid SCMP state', pygconst.Health.Warning),
    0xd: ('PECI over DMI failure', pygconst.Health.Warning),
    0xe: ('MCTP interface failure', pygconst.Health.Warning),
    0xf: ('Auto configuration completed', pygconst.Health.Ok),
}

me_flash_status = {
    0: ('ME flash corrupted', pygconst.Health.Critical),
    1: ('ME flash erase limit reached', pygconst.Health.Critical),
    2: ('ME flash write limit reached', pygconst.Health.Critical),
    3: ('ME flash write enabled', pygconst.Health.Ok),
}

leds = {
    "BMC_UID": 0x00,
    "BMC_HEARTBEAT": 0x01,
    "SYSTEM_FAULT": 0x02,
    "PSU1_FAULT": 0x03,
    "PSU2_FAULT": 0x04,
    "LED_FAN_FAULT_1": 0x10,
    "LED_FAN_FAULT_2": 0x11,
    "LED_FAN_FAULT_3": 0x12,
    "LED_FAN_FAULT_4": 0x13,
    "LED_FAN_FAULT_5": 0x14,
    "LED_FAN_FAULT_6": 0x15,
    "LED_FAN_FAULT_7": 0x16,
    "LED_FAN_FAULT_8": 0x17
}

led_status = {
    0x00: "Off",
    0xFF: "On"
}

ami_leds = {
    "BMC_HEARTBEAT": 0x00,
    "BMC_UID": 0x01,
    "SYSTEM_FAULT": 0x02,
    "HDD_FAULT": 0x03
}

asrock_leds = {
    "SYSTEM_EVENT": 0x00,
    "BMC_UID": 0x01,
    "LED_FAN_FAULT_1": 0x02,
    "LED_FAN_FAULT_2": 0x03,
    "LED_FAN_FAULT_3": 0x04
}

ts460_leds = {
    "SYSTEM_EVENT": 0x00,
    "BMC_UID": 0x01,
    "LED_FAN_FAULT_1": 0x02,
    "LED_FAN_FAULT_2": 0x03,
    "LED_FAN_CPU": 0x04,
    "LED_FAN_REAR": 0x05
}

asrock_led_status = {
    0x00: "Off",
    0x01: "On"
}

ami_led_status = {
    0x00: "Off",
    0x01: "On"
}

led_status_default = "Blink"
mac_format = '{0:02x}:{1:02x}:{2:02x}:{3:02x}:{4:02x}:{5:02x}'
categorie_items = ["cpu", "dimm", "firmware", "bios_version"]


def _megarac_abbrev_image(name):
    # MegaRAC platform in some places needs an abbreviated filename
    # Their scheme in such a scenario is a max of 20.  Truncation is
    # acheived by taking the first sixteen, then skipping ahead to the last
    # 4 (presumably to try to keep '.iso' or '.img' in the name).
    if len(name) <= 20:
        return name
    return name[:16] + name[-4:]


class OEMHandler(generic.OEMHandler):
    # noinspection PyUnusedLocal
    def __init__(self, oemid, ipmicmd):
        # will need to retain data to differentiate
        # variations.  For example System X versus Thinkserver
        self.oemid = oemid
        self._fpc_variant = None
        self.ipmicmd = weakref.proxy(ipmicmd)
        self._has_megarac = None
        self.oem_inventory_info = None
        self._mrethidx = None
        self._hasimm = None
        self._hasxcc = None
        if self.has_xcc:
            self.immhandler = imm.XCCClient(ipmicmd)
        elif self.has_imm:
            self.immhandler = imm.IMMClient(ipmicmd)
        elif self.is_fpc:
            self.smmhandler = nextscale.SMMClient(ipmicmd, self.is_fpc)
        elif self.has_tsma:
            conn = wc.SecureHTTPConnection(
                ipmicmd.bmc, 443,
                verifycallback=self.ipmicmd.certverify)
            self.tsmahandler = tsma.TsmHandler(None, None, conn,
                                               fish=redfishcmd)
            self.tsmahandler.set_credentials(
                ipmicmd.ipmi_session.userid.decode('utf-8'),
                ipmicmd.ipmi_session.password.decode('utf-8'))

    @property
    def _megarac_eth_index(self):
        if self._mrethidx is None:
            chan = self.ipmicmd.get_network_channel()
            rsp = self.ipmicmd.xraw_command(0x32, command=0x62, data=(chan,))
            self._mrethidx = rsp['data'][0]
        return self._mrethidx

    def remove_storage_configuration(self, cfgspec):
        if self.has_xcc:
            return self.immhandler.remove_storage_configuration(cfgspec)
        return super(OEMHandler, self).remove_storage_configuration()

    def clear_storage_arrays(self):
        if self.has_xcc:
            return self.immhandler.clear_storage_arrays()
        return super(OEMHandler, self).clear_storage_ararys()

    def apply_storage_configuration(self, cfgspec):
        if self.has_xcc:
            return self.immhandler.apply_storage_configuration(cfgspec)
        return super(OEMHandler, self).apply_storage_configuration()

    def check_storage_configuration(self, cfgspec):
        if self.has_xcc:
            return self.immhandler.check_storage_configuration(cfgspec)
        return super(OEMHandler, self).get_storage_configuration()

    def get_storage_configuration(self):
        if self.has_xcc:
            return self.immhandler.get_storage_configuration()
        return super(OEMHandler, self).get_storage_configuration()

    def get_video_launchdata(self):
        if self.has_tsm:
            return self.get_tsm_launchdata()

    def get_tsm_launchdata(self):
        pass

    def process_event(self, event, ipmicmd, seldata):
        if 'oemdata' in event:
            oemtype = seldata[2]
            oemdata = event['oemdata']
            if oemtype == 0xd0:  # firmware update
                event['component'] = firmware_types.get(oemdata[0], None)
                event['component_type'] = ipmiconst.sensor_type_codes[0x2b]
                slotnumber = (oemdata[1] & 0b11111000) >> 3
                if slotnumber:
                    event['component'] += ' {0}'.format(slotnumber)
                event['event'], event['severity'] = \
                    firmware_event[oemdata[1] & 0b111]
                event['event_data'] = '{0}.{1}'.format(oemdata[2], oemdata[3])
            elif oemtype == 0xd1:  # BIOS recovery
                event['severity'] = pygconst.Health.Warning
                event['component'] = 'BIOS/UEFI'
                event['component_type'] = ipmiconst.sensor_type_codes[0xf]
                status = oemdata[0]
                method = (status & 0b11110000) >> 4
                status = (status & 0b1111)
                if method == 1:
                    event['event'] = 'Automatic recovery'
                elif method == 2:
                    event['event'] = 'Manual recovery'
                if status == 0:
                    event['event'] += '- Failed'
                    event['severity'] = pygconst.Health.Failed
                if oemdata[1] == 0x1:
                    event['event'] += '- BIOS recovery image not found'
                event['event_data'] = '{0}.{1}'.format(oemdata[2], oemdata[3])
            elif oemtype == 0xd2:  # eMMC status
                if oemdata[0] == 1:
                    event['component'] = 'eMMC'
                    event['component_type'] = ipmiconst.sensor_type_codes[0xc]
                    if oemdata[0] == 1:
                        event['event'] = 'eMMC Format error'
                        event['severity'] = pygconst.Health.Failed
            elif oemtype == 0xd3:
                if oemdata[0] == 1:
                    event['event'] = 'User privilege modification'
                    event['severity'] = pygconst.Health.Ok
                    event['component'] = 'User Privilege'
                    event['component_type'] = ipmiconst.sensor_type_codes[6]
                    event['event_data'] = (
                        'User {0} on channel {1} had privilege changed '
                        'from {2} to {3}'.format(
                            oemdata[2], oemdata[1], oemdata[3] & 0b1111,
                            (oemdata[3] & 0b11110000) >> 4)
                    )
            else:
                event['event'] = 'OEM event: {0}'.format(
                    ' '.join(format(x, '02x') for x in event['oemdata']))
            del event['oemdata']
            return
        evdata = event['event_data_bytes']
        if event['event_type_byte'] == 0x75:  # ME event
            event['component'] = 'ME Firmware'
            event['component_type'] = ipmiconst.sensor_type_codes[0xf]
            event['event'], event['severity'] = me_status.get(
                evdata[1], ('Unknown', pygconst.Health.Warning))
            if evdata[1] == 3:
                event['event'], event['severity'] = me_flash_status.get(
                    evdata[2], ('Unknown state', pygconst.Health.Warning))
            elif evdata[1] == 9:
                event['event'] += ' (0x{0:2x})'.format(evdata[2])
            elif evdata[1] == 0xf and evdata[2] & 0b10000000:
                event['event'] = 'Auto configuration failed'
                event['severity'] = pygconst.Health.Critical
        # For HDD bay events, the event data 2 is the bay, modify
        # the description to be more specific
        if (event['event_type_byte'] == 0x6f
                and (evdata[0] & 0b11000000) == 0b10000000
                and event['component_type_id'] == 13):
            event['component'] += ' {0}'.format(evdata[1] & 0b11111)

    def reseat_bay(self, bay):
        if self.is_fpc:
            return self.smmhandler.reseat_bay(bay)
        elif self.has_xcc and bay == -1:
            return self.immhandler.reseat()
        return super(OEMHandler, self).reseat_bay(bay)

    def get_ntp_enabled(self):
        if self.has_tsm or self.has_ami or self.has_asrock:
            ntpres = self.ipmicmd.xraw_command(netfn=0x32, command=0xa7)
            return ntpres['data'][0] == '\x01'
        elif self.is_fpc:
            return self.smmhandler.get_ntp_enabled(self._fpc_variant)
        elif self.has_tsma:
            return self.tsmahandler.get_ntp_enabled()
        return None

    def get_ntp_servers(self):
        if self.has_tsm or self.has_ami or self.has_asrock:
            srvs = []
            ntpres = self.ipmicmd.xraw_command(netfn=0x32, command=0xa7)
            srvs.append(ntpres['data'][1:129].rstrip('\x00'))
            srvs.append(ntpres['data'][129:257].rstrip('\x00'))
            return srvs
        if self.is_fpc:
            return self.smmhandler.get_ntp_servers()
        if self.has_tsma:
            return self.tsmahandler.get_ntp_servers()
        return None

    def set_ntp_enabled(self, enabled):
        if self.has_tsm or self.has_ami or self.has_asrock:
            if enabled:
                self.ipmicmd.xraw_command(
                    netfn=0x32, command=0xa8, data=(3, 1), timeout=15)
            else:
                self.ipmicmd.xraw_command(
                    netfn=0x32, command=0xa8, data=(3, 0), timeout=15)
            return True
        if self.is_fpc:
            self.smmhandler.set_ntp_enabled(enabled)
            return True
        if self.has_tsma:
            self.tsmahandler.set_ntp_enabled(enabled)
        return None

    def set_ntp_server(self, server, index=0):
        if self.has_tsm or self.has_ami or self.has_asrock:
            if not (0 <= index <= 1):
                raise pygexc.InvalidParameterValue("Index must be 0 or 1")
            cmddata = bytearray((1 + index, ))
            cmddata += server.ljust(128, '\x00')
            self.ipmicmd.xraw_command(netfn=0x32, command=0xa8, data=cmddata)
            return True
        elif self.is_fpc:
            if not 0 <= index <= 2:
                raise pygexc.InvalidParameterValue(
                    'SMM supports indexes 0 through 2')
            self.smmhandler.set_ntp_server(server, index)
            return True
        elif self.has_tsma:
            if not (0 <= index <= 1):
                raise pygexc.InvalidParameterValue("Index must be 0 or 1")
            return self.tsmahandler.set_ntp_server(server, index)
        return None
    
    def user_delete(self, uid, channel):
        if self.has_xcc:
            return self.immhandler.user_delete(uid)

    def set_user_access(self, uid, channel, callback, link_auth, ipmi_msg,
                        privilege_level):
        if self.is_fpc and self._fpc_variant != 6:
            self.smmhandler.set_user_priv(uid, privilege_level)

    @property
    def is_fpc(self):
        """True if the target is a Lenovo nextscale fan power controller"""

        if self.has_imm or self.has_xcc:
            return None
        if self._fpc_variant is not None:
            return self._fpc_variant
        fpc_ids = ((19046, 32, 1063), (20301, 32, 462))
        smm_id = (19046, 32, 1180)
        smmv2_2u = (19046, 32, 1360)
        smmv2_6u = (19046, 32, 1387)
        currid = (self.oemid['manufacturer_id'], self.oemid['device_id'],
                  self.oemid['product_id'])
        if currid in fpc_ids:
            self._fpc_variant = 6
        elif currid == smm_id:
            self._fpc_variant = 2
        elif currid == smmv2_2u:
            self._fpc_variant = 0x22
        elif currid == smmv2_6u:
            self._fpc_variant = 0x26
        return self._fpc_variant

    @property
    def is_sd350(self):
        return (19046, 32, 13616) == (self.oemid['manufacturer_id'],
                                      self.oemid['device_id'],
                                      self.oemid['product_id'])

    tsma_ids = ((19046, 32, 1287),)

    @property
    def has_tsma(self):
        currid = (self.oemid['manufacturer_id'], self.oemid['device_id'],
                  self.oemid['product_id'])
        return currid in self.tsma_ids

    @property
    def has_tsm(self):
        """True if this particular server have a TSM based service processor"""

        if (self.oemid['manufacturer_id'] == 19046
                and self.oemid['device_id'] == 32):
            try:
                self.ipmicmd.xraw_command(netfn=0x3a, command=0xf)
            except pygexc.IpmiException as ie:
                if ie.ipmicode == 193:
                    return False
                raise
            return True
        return False

    @property
    def has_asrock(self):
        # True if this particular server have a ASROCKRACK
        # based service processor (RS160 or TS460)
        # RS160 (Riddler) product id is 1182 (049Eh)
        # TS460 (WildThing) product id is 1184 (04A0h)

        if (self.oemid['manufacturer_id'] == 19046
                and (self.oemid['product_id'] == 1182
                     or self.oemid['product_id'] == 1184)):
            try:
                self.ipmicmd.xraw_command(netfn=0x3a,
                                          command=0x50,
                                          data=(0x00, 0x00, 0x00))
            except pygexc.IpmiException as ie:
                if ie.ipmicode == 193:
                    return False
                raise
            return True
        return False

    @property
    def isTS460(self):
        if self.oemid['product_id'] == 1184:
            return True
        return False

    def get_oem_inventory_descriptions(self):
        if self.has_tsm or self.has_ami or self.has_asrock:
            # Thinkserver with TSM
            if not self.oem_inventory_info:
                self._collect_tsm_inventory()
            return iter(self.oem_inventory_info)
        elif self.has_imm:
            return self.immhandler.get_hw_descriptions()
        elif self.is_fpc:
            return self.smmhandler.get_inventory_descriptions(self.ipmicmd,
                                                              self.is_fpc)
        return ()

    def get_oem_inventory(self):
        if self.has_tsm or self.has_ami or self.has_asrock:
            self._collect_tsm_inventory()
            for compname in self.oem_inventory_info:
                yield (compname, self.oem_inventory_info[compname])
        elif self.has_imm:
            for inv in self.immhandler.get_hw_inventory():
                yield inv
        elif self.is_fpc:
            for compname in self.smmhandler.get_inventory_descriptions(
                    self.ipmicmd, self.is_fpc):
                yield (compname, self.smmhandler.get_inventory_of_component(
                    self.ipmicmd, compname))

    def get_sensor_data(self):
        if self.has_imm:
            for name in self.immhandler.get_oem_sensor_names(self.ipmicmd):
                yield self.immhandler.get_oem_sensor_reading(name,
                                                             self.ipmicmd)
        elif self.is_fpc:
            for name in nextscale.get_sensor_names(self.ipmicmd,
                                                   self._fpc_variant):
                yield nextscale.get_sensor_reading(name, self.ipmicmd,
                                                   self._fpc_variant)
        elif self.has_ami:
            self.get_ami_sensor_data()

    def get_sensor_descriptions(self):
        if self.has_imm:
            return self.immhandler.get_oem_sensor_descriptions(self.ipmicmd)
        elif self.is_fpc:
            return nextscale.get_sensor_descriptions(
                self.ipmicmd, self._fpc_variant)
        elif self.has_ami:
            self.get_ami_sensor_descriptions()
        return ()

    def get_sensor_reading(self, sensorname):
        if self.has_imm:
            return self.immhandler.get_oem_sensor_reading(sensorname,
                                                          self.ipmicmd)
        elif self.is_fpc:
            return nextscale.get_sensor_reading(sensorname, self.ipmicmd,
                                                self._fpc_variant)
        elif self.has_ami:
            self.get_ami_sensor_reading(sensorname)
        return ()

    def get_inventory_of_component(self, component):
        if self.has_tsm or self.has_ami or self.has_asrock:
            self._collect_tsm_inventory()
            return self.oem_inventory_info.get(component, None)
        if self.has_imm:
            return self.immhandler.get_component_inventory(component)
        if self.is_fpc:
            return self.smmhandler.get_inventory_of_component(component)

    def get_cmd_type(self, categorie_item, catspec):
        if self.has_asrock:
            cmd_type = catspec["command"]["asrock"]
        elif categorie_item in categorie_items:
            cmd_type = catspec["command"]["lenovo"]
        else:
            cmd_type = catspec["command"]

        return cmd_type

    def _collect_tsm_inventory(self):
        self.oem_inventory_info = {}
        asrock = False
        if self.has_asrock:
            asrock = True
        for catid, catspec in inventory.categories.items():
            # skip the inventory fields if the system is RS160
            if asrock and catid not in categorie_items:
                continue
            if (catspec.get("workaround_bmc_bug", False)
                    and catspec["workaround_bmc_bug"](
                        "ami" if self.has_ami else "lenovo")):
                rsp = None
                cmd = self.get_cmd_type(catid, catspec)
                tmp_command = dict(cmd)
                tmp_command["data"] = list(tmp_command["data"])
                count = 0
                for i in range(0x01, 0xff):
                    tmp_command["data"][-1] = i
                    try:
                        partrsp = self.ipmicmd.xraw_command(**tmp_command)
                        count += 1
                        if asrock and partrsp["data"][1] == "\xff":
                            continue

                        if rsp is None:
                            rsp = partrsp
                            rsp["data"] = list(rsp["data"])
                        else:
                            rsp["data"].extend(partrsp["data"][1:])
                    except Exception:
                        break
                # If we didn't get any response, assume we don't have
                # this category and go on to the next one
                if rsp is None:
                    continue
                rsp["data"].insert(1, count)
                rsp["data"] = buffer(bytearray(rsp["data"]))
            else:
                try:
                    cmd = self.get_cmd_type(catid, catspec)
                    rsp = self.ipmicmd.xraw_command(**cmd)
                except pygexc.IpmiException:
                    continue
            # Parse the response we got
            try:
                items = inventory.parse_inventory_category(
                    catid, rsp, asrock,
                    countable=catspec.get("countable", True)
                )
            except Exception:
                # If we can't parse an inventory category, ignore it
                print(traceback.print_exc())
                continue

            for item in items:
                try:
                    # Originally on ThinkServer and SD350 (Kent),
                    # the DIMM is distinguished by slot,
                    # the key is the value of slot number (item["index"])
                    # While on RS160/TS460 the DIMMs is distinguished
                    # by slot number and channel number,
                    # the key is the value of the sum of slot number
                    # and channel number
                    if asrock and catid == "dimm":
                        if item["channel_number"] == 1:
                            key = catspec["idstr"].format(item["index"])
                        else:
                            key = catspec["idstr"].format(
                                item["index"] + item["channel_number"])
                    else:
                        key = catspec["idstr"].format(item["index"])

                    del item["index"]
                    self.oem_inventory_info[key] = item
                except Exception:
                    # If we can't parse an inventory item, ignore it
                    print(traceback.print_exc())
                    continue

    def get_leds(self):
        cmd = 0x02
        led_set = leds
        led_set_status = led_status

        asrock = self.has_asrock
        if self.has_ami:
            cmd = 0x05
            led_set = ami_leds
            led_set_status = ami_led_status
        elif asrock:
            cmd = 0x50
            # because rs160 has different led info with ts460
            if self.isTS460:
                led_set = ts460_leds
            else:
                led_set = asrock_leds
            led_set_status = asrock_led_status

        for (name, id_) in led_set.items():
            try:
                if asrock:
                    rsp = self.ipmicmd.xraw_command(netfn=0x3A, command=cmd,
                                                    data=(0x03, id_, 0x00))
                    rdata = bytearray(rsp['data'][:])
                    status = rdata[1]
                else:
                    rsp = self.ipmicmd.xraw_command(netfn=0x3A, command=cmd,
                                                    data=(id_,))
                    rdata = bytearray(rsp['data'][:])
                    status = rdata[0]

            except pygexc.IpmiException:
                continue  # Ignore LEDs we can't retrieve
            status = led_set_status.get(status,
                                        led_status_default)
            yield (name, {'status': status})

    def set_identify(self, on, duration):
        if on and not duration and self.is_sd350:
            self.ipmicmd.xraw_command(netfn=0x3a, command=6, data=(1, 1))
        else:
            raise pygexc.UnsupportedFunctionality()

    def process_fru(self, fru, name=None):
        if fru is None:
            return fru
        if self.has_tsm:
            fru['oem_parser'] = 'lenovo'
            # Thinkserver lays out specific interpretation of the
            # board extra fields
            try:
                _, _, wwn1, wwn2, mac1, mac2 = fru['board_extra']
                if wwn1 not in ('0000000000000000', ''):
                    fru['WWN 1'] = wwn1.encode('utf-8')
                if wwn2 not in ('0000000000000000', ''):
                    fru['WWN 2'] = wwn2.encode('utf-8')
                if mac1 not in ('00:00:00:00:00:00', ''):
                    fru['MAC Address 1'] = mac1.encode('utf-8')
                if mac2 not in ('00:00:00:00:00:00', ''):
                    fru['MAC Address 2'] = mac2.encode('utf-8')
            except (AttributeError, KeyError):
                pass
            try:
                # The product_extra field is UUID as the system would present
                # in DMI.  This is different than the two UUIDs that
                # it returns for get device and get system uuid...
                byteguid = fru['product_extra'][0]
                # It can present itself as claiming to be ASCII when it
                # is actually raw hex.  As a result it triggers the mechanism
                # to strip \x00 from the end of text strings.  Work around this
                # by padding with \x00 to the right if less than 16 long
                byteguid.extend('\x00' * (16 - len(byteguid)))
                if byteguid not in ('\x20' * 16, '\x00' * 16, '\xff' * 16):
                    fru['UUID'] = util.decode_wireformat_uuid(byteguid)
            except (AttributeError, KeyError, IndexError):
                pass
            return fru
        elif self.has_imm:
            fru['oem_parser'] = 'lenovo'
            try:
                bextra = fru['board_extra']
                fru['FRU Number'] = bextra[0]
                fru['Revision'] = bextra[4]
                macs = bextra[6]
                macprefix = None
                idx = 0
                endidx = len(macs) - 5
                macprefix = None
                while idx < endidx:
                    currmac = macs[idx:idx + 6]
                    if not isinstance(currmac, bytearray):
                        # invalid vpd format, abort attempts to extract
                        # mac in this way
                        break
                    if currmac == b'\x00\x00\x00\x00\x00\x00':
                        break
                    # VPD may veer off, detect and break off
                    if macprefix is None:
                        macprefix = currmac[:3]
                    elif currmac[:3] != macprefix:
                        break
                    ms = mac_format.format(*currmac)
                    ifidx = idx / 6 + 1
                    fru['MAC Address {0}'.format(ifidx)] = ms
                    idx = idx + 6
            except (AttributeError, KeyError, IndexError):
                pass
            if self.has_xcc and name and name.startswith('PSU '):
                self.immhandler.augment_psu_info(fru, name)
            if (self.has_xcc and 'memory_type' in fru
                    and fru['memory_type'] == 'Unknown'):
                self.immhandler.fetch_dimm(name, fru)
            return fru
        elif self.is_fpc and self.is_fpc != 6:  # SMM variant
            fru['oem_parser'] = 'lenovo'
            return self.smmhandler.process_fru(fru)
        elif self.has_asrock:
            fru['oem_parser'] = 'lenovo'
            # ASRock RS160 TS460 lays out specific interpretation of the
            # board extra fields
            try:
                mac1 = fru['board_extra']
                if mac1 not in ('00:00:00:00:00:00', ''):
                    fru['MAC Address 1'] = mac1.encode('utf-8')
            except (AttributeError, KeyError):
                pass
            return fru
        else:
            fru['oem_parser'] = None
            return fru

    @property
    def has_xcc(self):
        if self._hasxcc is not None:
            return self._hasxcc
        try:
            bdata = self.ipmicmd.xraw_command(netfn=0x3a, command=0xc1)
        except pygexc.IpmiException:
            self._hasxcc = False
            self._hasimm = False
            return False
        if len(bdata['data'][:]) != 3:
            self._hasimm = False
            self._hasxcc = False
            return False
        rdata = bytearray(bdata['data'][:])
        self._hasxcc = rdata[1] & 16 == 16
        if self._hasxcc:
            # For now, have imm calls go to xcc, since they are providing same
            # interface.  Longer term the hope is that all the Lenovo
            # stuff will branch at init, and not have conditionals
            # in all the functions
            self._hasimm = self._hasxcc
        return self._hasxcc

    @property
    def has_imm(self):
        if self._hasimm is not None:
            return self._hasimm
        try:
            bdata = self.ipmicmd.xraw_command(netfn=0x3a, command=0xc1)
        except pygexc.IpmiException:
            self._hasimm = False
            return False
        if len(bdata['data'][:]) != 3:
            self._hasimm = False
            return False
        rdata = bytearray(bdata['data'][:])
        self._hasimm = (rdata[1] & 1 == 1) or (rdata[1] & 16 == 16)
        return self._hasimm

    @property
    def has_ami(self):
        """True if this particular server is AMI based lenovo server

        """
        if(self.oemid['manufacturer_id'] == 19046
                and self.oemid['product_id'] == 13616):
            try:
                rsp = self.ipmicmd.xraw_command(netfn=0x3a, command=0x80)
            except pygexc.IpmiException as ie:
                if ie.ipmicode == 193:
                    return False
                raise
            rdata = bytearray(rsp['data'][:])
            if rdata[0] in range(5):
                return True
            else:
                return False
        return False

    def get_oem_firmware(self, bmcver, components):
        if self.has_tsm or self.has_ami or self.has_asrock:
            command = firmware.get_categories()["firmware"]
            fw_cmd = self.get_cmd_type("firmware", command)
            rsp = self.ipmicmd.xraw_command(**fw_cmd)

            # the newest Lenovo ThinkServer versions are returning Bios version
            # numbers through another command
            bios_versions = None
            if self.has_tsm or self.has_asrock:
                bios_command = firmware.get_categories()["bios_version"]
                bios_cmd = self.get_cmd_type("bios_version", bios_command)
                bios_rsp = self.ipmicmd.xraw_command(**bios_cmd)
                if self.has_asrock:
                    bios_versions = bios_rsp['data']
                else:
                    bios_versions = bios_command["parser"](bios_rsp['data'])

            # pass bios versions to firmware parser
            return command["parser"](rsp["data"],
                                     bios_versions,
                                     self.has_asrock)
        elif self.has_imm:
            return self.immhandler.get_firmware_inventory(bmcver, components)
        elif self.is_fpc:
            return nextscale.get_fpc_firmware(bmcver, self.ipmicmd,
                                              self._fpc_variant)
        elif self.has_tsma:
            return self.tsmahandler.get_firmware_inventory(
                components, raisebypass=False, ipmicmd=self.ipmicmd)
        return super(OEMHandler, self).get_oem_firmware(bmcver, components)

    def get_diagnostic_data(self, savefile, progress, autosuffix=False):
        if self.has_xcc:
            return self.immhandler.get_diagnostic_data(savefile, progress,
                                                       autosuffix)
        if self.is_fpc:
            return self.smmhandler.get_diagnostic_data(savefile, progress,
                                                       autosuffix,
                                                       self._fpc_variant)
        if self.has_tsma:
            return self.tsmahandler.get_diagnostic_data(savefile, progress,
                                                        autosuffix)

    def get_oem_capping_enabled(self):
        if self.has_tsm:
            rsp = self.ipmicmd.xraw_command(netfn=0x3a, command=0x1b,
                                            data=(3,))
            # disabled
            if rsp['data'][0] == '\x00':
                return False
            # enabled
            else:
                return True

    def set_oem_capping_enabled(self, enable):
        """Set PSU based power capping

        :param enable: True for enable and False for disable
        """
        # 1 - Enable power capping(default)
        if enable:
            statecode = 1
        # 0 - Disable power capping
        else:
            statecode = 0
        if self.has_tsm:
            self.ipmicmd.xraw_command(netfn=0x3a, command=0x1a,
                                      data=(3, statecode))
            return True

    def get_oem_remote_kvm_available(self):
        if self.has_tsm:
            rsp = self.ipmicmd.raw_command(netfn=0x3a, command=0x13)
            return rsp['data'][0] == 0
        return False

    def _restart_dns(self):
        if self.has_tsm:
            self.ipmicmd.xraw_command(netfn=0x32, command=0x6c, data=(7, 0))

    def get_oem_domain_name(self):
        if self.has_tsm:
            name = ''
            for i in range(1, 5):
                rsp = self.ipmicmd.xraw_command(netfn=0x32, command=0x6b,
                                                data=(4, i))
                name += rsp['data'][:]
            return name.rstrip('\x00')
        elif self.is_fpc:
            return self.smmhandler.get_domain()

    def set_oem_domain_name(self, name):
        if self.has_tsm:
            # set the domain name length
            data = [3, 0, 0, 0, 0, len(name)]
            self.ipmicmd.xraw_command(netfn=0x32, command=0x6c, data=data)

            # set the domain name content
            name = name.ljust(256, "\x00")
            for i in range(0, 4):
                data = [4, i + 1]
                offset = i * 64
                data.extend([ord(x) for x in name[offset:offset + 64]])
                self.ipmicmd.xraw_command(netfn=0x32, command=0x6c, data=data)

            self._restart_dns()
            return
        elif self.is_fpc:
            self.smmhandler.set_domain(name)

    def set_hostname(self, hostname):
        if self.has_xcc:
            return self.immhandler.set_hostname(hostname)
        elif self.is_fpc:
            return self.smmhandler.set_hostname(hostname)
        return super(OEMHandler, self).set_hostname(hostname)

    def get_hostname(self):
        if self.has_xcc:
            return self.immhandler.get_hostname()
        elif self.is_fpc:
            return self.smmhandler.get_hostname()
        return super(OEMHandler, self).get_hostname()

    """ Gets a remote console launcher for a Lenovo ThinkServer.

    Returns a tuple: (content type, launcher) or None if the launcher could
    not be retrieved."""
    def _get_ts_remote_console(self, bmc, username, password):
        # We don't establish non-secure connections without checking
        # certificates
        if not self.ipmicmd.certverify:
            return
        conn = wc.SecureHTTPConnection(bmc, 443,
                                       verifycallback=self.ipmicmd.certverify)
        conn.connect()
        params = urlencode({
            'WEBVAR_USERNAME': username,
            'WEBVAR_PASSWORD': password
        })
        headers = {
            'Connection': 'keep-alive'
        }
        conn.request('POST', '/rpc/WEBSES/create.asp', params, headers)
        rsp = conn.getresponse()
        if rsp.status == 200:
            conn.cookies = {}
            body = rsp.read().split('\n')
            session_line = None
            for line in body:
                if 'SESSION_COOKIE' in line:
                    session_line = line
            if session_line is None:
                return

            session_id = session_line.split('\'')[3]
            # Usually happens when maximum number of sessions is reached
            if session_id == 'Failure_Session_Creation':
                return

            headers = {
                'Connection': 'keep-alive',
                'Cookie': 'SessionCookie=' + session_id,
            }
            conn.request(
                'GET',
                '/Java/jviewer.jnlp?EXTRNIP=' + bmc + '&JNLPSTR=JViewer',
                None, headers)
            rsp = conn.getresponse()
            if rsp.status == 200:
                return rsp.getheader('Content-Type'), base64.b64encode(
                    rsp.read())
        conn.close()

    def get_graphical_console(self):
        return self._get_ts_remote_console(self.ipmicmd.bmc,
                                           self.ipmicmd.ipmi_session.userid,
                                           self.ipmicmd.ipmi_session.password)

    def add_extra_net_configuration(self, netdata, channel=None):
        if self.has_tsm:
            ipv6_addr = self.ipmicmd.xraw_command(
                netfn=0x0c, command=0x02,
                data=(0x01, 0xc5, 0x00, 0x00))["data"][1:]
            if not ipv6_addr:
                return
            rspdata = self.ipmicmd.xraw_command(
                netfn=0xc, command=0x02,
                data=(0x1, 0xc6, 0, 0))['data']
            ipv6_prefix_ba = bytearray(rspdata)
            ipv6_prefix = ipv6_prefix_ba[1]

            if hasattr(socket, 'inet_ntop'):
                ipv6str = socket.inet_ntop(socket.AF_INET6, ipv6_addr)
            else:
                # fall back to a dumber, but more universal formatter
                ipv6str = binascii.b2a_hex(ipv6_addr)
                ipv6str = ':'.join([ipv6str[x:x + 4] for x in range(0, 32, 4)])
            netdata['ipv6_addresses'] = [
                '{0}/{1}'.format(ipv6str, ipv6_prefix)]

    @property
    def has_megarac(self):
        # if there is functionality that is the same for tsm or generic
        # megarac, then this is appropriate.  If there's a TSM specific
        # preferred, use has_tsm first
        if self._has_megarac is not None:
            return self._has_megarac
        self._has_megarac = False
        try:
            rsp = self.ipmicmd.xraw_command(netfn=0x32, command=0x7e)
            # We don't have a handy classify-only, so use get sel policy
            # rsp should have a length of one, and be either '\x00' or '\x01'
            if len(rsp['data'][:]) == 1 and rsp['data'][0] in ('\x00', '\x01'):
                self._has_megarac = True
        except pygexc.IpmiException as ie:
            if ie.ipmicode == 0:
                # if it's a generic IpmiException rather than an error code
                # from the BMC, then this is a deeper problem than just an
                # invalid command or command length or similar
                raise
        return self._has_megarac

    def set_alert_ipv6_destination(self, ip, destination, channel):
        if self.has_megarac:
            ethidx = self._megarac_eth_index
            reqdata = bytearray([channel, 193, destination, ethidx, 0])
            parsedip = socket.inet_pton(socket.AF_INET6, ip)
            reqdata.extend(parsedip)
            reqdata.extend('\x00\x00\x00\x00\x00\x00')
            self.ipmicmd.xraw_command(netfn=0xc, command=1, data=reqdata)
            return True
        return False

    def _set_short_ris_string(self, selector, value):
        data = (1, selector, 0) + struct.unpack('{0}B'.format(len(value)),
                                                value)
        self.ipmicmd.xraw_command(netfn=0x32, command=0x9f, data=data)

    def _set_ris_string(self, selector, value):
        if len(value) > 256:
            raise pygexc.UnsupportedFunctionality(
                'Value exceeds 256 characters: {0}'.format(value))
        padded = value + (256 - len(value)) * '\x00'
        padded = list(struct.unpack('256B', padded))
        # 8 = RIS, 4 = hd, 2 = fd, 1 = cd
        try:  # try and clear in-progress if left incomplete
            self.ipmicmd.xraw_command(netfn=0x32, command=0x9f,
                                      data=(1, selector, 0, 0))
        except pygexc.IpmiException:
            pass
        # set in-progress
        self.ipmicmd.xraw_command(netfn=0x32, command=0x9f,
                                  data=(1, selector, 0, 1))
        # now do the set
        for x in range(0, 256, 64):
            currdata = padded[x:x + 64]
            currchunk = x // 64 + 1
            cmddata = [1, selector, currchunk] + currdata
            self.ipmicmd.xraw_command(netfn=0x32, command=0x9f, data=cmddata)
        # unset in-progress
        self.ipmicmd.xraw_command(netfn=0x32, command=0x9f,
                                  data=(1, selector, 0, 0))

    def _megarac_fetch_image_shortnames(self):
        rsp = self.ipmicmd.xraw_command(netfn=0x32, command=0xd8,
                                        data=(7, 1, 0))
        imgnames = rsp['data'][1:]
        shortnames = []
        for idx in range(0, len(imgnames), 22):
            shortnames.append(imgnames[idx + 2:idx + 22].rstrip('\0'))
        return shortnames

    def _megarac_media_waitforready(self, imagename):
        # first, we have, sadly, a 10 second grace period for some invisible
        # async activity to get far enough long to monitor
        self.ipmicmd.ipmi_session.pause(10)
        risenabled = '\x00'
        mountok = '\xff'
        while risenabled != '\x01':
            risenabled = self.ipmicmd.xraw_command(
                netfn=0x32, command=0x9e, data=(8, 10))['data'][2]
        while mountok == '\xff':
            mountok = self.ipmicmd.xraw_command(
                netfn=0x32, command=0x9e, data=(1, 8))['data'][2]
        targshortname = _megarac_abbrev_image(imagename)
        shortnames = self._megarac_fetch_image_shortnames()
        while targshortname not in shortnames:
            self.ipmicmd.wait_for_rsp(1)
            shortnames = self._megarac_fetch_image_shortnames()
        self.ipmicmd.ipmi_session.pause(10)
        try:
            self.ipmicmd.xraw_command(netfn=0x32, command=0xa0, data=(1, 0))
            self.ipmicmd.ipmi_session.pause(5)
        except pygexc.IpmiException:
            pass

    def _megarac_attach_media(self, proto, username, password, imagename,
                              domain, path, host):
        # First we must ensure that the RIS is actually enabled
        self.ipmicmd.xraw_command(netfn=0x32, command=0x9f, data=(8, 10, 0, 1))
        if username is not None:
            self._set_ris_string(3, username)
        if password is not None:
            self._set_short_ris_string(4, password)
        if domain is not None:
            self._set_ris_string(6, domain)
        self._set_ris_string(1, path)
        ip = util.get_ipv4(host)[0]
        self._set_short_ris_string(2, ip)
        self._set_short_ris_string(5, proto)
        # now to restart RIS to have changes take effect...
        self.ipmicmd.xraw_command(netfn=0x32, command=0x9f, data=(8, 11))
        # now to kick off the requested mount
        self._megarac_media_waitforready(imagename)
        self._set_ris_string(0, imagename)
        self.ipmicmd.xraw_command(netfn=0x32, command=0xa0,
                                  data=(1, 1))

    def attach_remote_media(self, url, username, password):
        if self.has_imm:
            self.immhandler.attach_remote_media(url, username, password)
        elif self.has_tsma:
            return self.tsmahandler.attach_remote_media(
                url, username, password, None)
        elif self.has_megarac:
            proto, host, path = util.urlsplit(url)
            if proto == 'smb':
                proto = 'cifs'
            domain = None
            path, imagename = path.rsplit('/', 1)
            if username is not None and '@' in username:
                username, domain = username.split('@', 1)
            elif username is not None and '\\' in username:
                domain, username = username.split('\\', 1)
            try:
                self._megarac_attach_media(proto, username, password,
                                           imagename, domain, path, host)
            except pygexc.IpmiException as ie:
                if ie.ipmicode in (0x92, 0x99):
                    # if starting from scratch, this can happen...
                    self._megarac_attach_media(proto, username, password,
                                               imagename, domain, path, host)
                else:
                    raise

    def update_firmware(self, filename, data=None, progress=None, bank=None):
        if self.has_xcc:
            return self.immhandler.update_firmware(
                filename, data=data, progress=progress, bank=bank)
        if self.is_fpc:
            return self.smmhandler.update_firmware(
                filename, data=data, progress=progress, bank=bank)
        if self.has_tsma:
            return self.tsmahandler.update_firmware(
                filename, data=data, progress=progress, bank=bank)
        return super(OEMHandler, self).update_firmware(filename, data=data,
                                                       progress=progress,
                                                       bank=bank)

    def get_description(self):
        if self.has_xcc:
            return self.immhandler.get_description()
        if self.is_fpc:
            return {'height': self._fpc_variant, 'slot': 0}
        return super(OEMHandler, self).get_description()

    def get_extended_bmc_configuration(self):
        if self.has_xcc:
            return self.immhandler.get_extended_bmc_configuration()
        return super(OEMHandler, self).get_extended_bmc_configuration()

    def get_bmc_configuration(self):
        if self.has_xcc:
            return self.immhandler.get_bmc_configuration()
        if self.is_fpc:
            return self.smmhandler.get_bmc_configuration()
        if self.has_tsma:
            return self.tsmahandler.get_bmc_configuration()
        return super(OEMHandler, self).get_bmc_configuration()

    def set_bmc_configuration(self, changeset):
        if self.has_xcc:
            return self.immhandler.set_bmc_configuration(changeset)
        if self.is_fpc:
            return self.smmhandler.set_bmc_configuration(
                changeset, self._fpc_variant)
        if self.has_tsma:
            return self.tsmahandler.set_bmc_configuration(
                changeset)
        return super(OEMHandler, self).set_bmc_configuration(changeset)

    def get_system_configuration(self, hideadvanced):
        if self.has_imm or self.has_xcc:
            return self.immhandler.get_system_configuration(hideadvanced)
        if self.has_tsma:
            return self.tsmahandler.get_uefi_configuration(hideadvanced)
        return super(OEMHandler, self).get_system_configuration(hideadvanced)

    def set_system_configuration(self, changeset):
        if self.has_imm or self.has_xcc:
            return self.immhandler.set_system_configuration(changeset)
        if self.has_tsma:
            return self.tsmahandler.set_uefi_configuration(changeset)
        return super(OEMHandler, self).set_system_configuration(changeset)

    def clear_bmc_configuration(self):
        if self.has_xcc:
            return self.immhandler.clear_bmc_configuration()
        elif self.is_fpc:
            return self.smmhandler.clear_bmc_configuration()
        elif self.has_tsma:
            return self.tsmahandler.clear_bmc_configuration()
        return super(OEMHandler, self).clear_system_configuration()

    def clear_system_configuration(self):
        if self.has_xcc:
            return self.immhandler.clear_system_configuration()
        if self.has_tsma:
            return self.tsmahandler.clear_uefi_configuration()
        return super(OEMHandler, self).clear_system_configuration()

    def detach_remote_media(self):
        if self.has_imm:
            self.immhandler.detach_remote_media()
        elif self.has_tsma:
            self.tsmahandler.detach_remote_media()
        elif self.has_megarac:
            self.ipmicmd.xraw_command(
                netfn=0x32, command=0x9f, data=(8, 10, 0, 0))
            self.ipmicmd.xraw_command(netfn=0x32, command=0x9f, data=(8, 11))

    def upload_media(self, filename, progress, data):
        if self.has_xcc or self.has_imm:
            return self.immhandler.upload_media(filename, progress, data)
        return super(OEMHandler, self).upload_media(filename, progress, data)

    def list_media(self):
        if self.has_xcc or self.has_imm:
            return self.immhandler.list_media()
        if self.has_tsma:
            return self.tsmahandler.list_media()
        return super(OEMHandler, self).list_media()

    def get_health(self, summary):
        if self.has_xcc:
            return self.immhandler.get_health(summary)
        return super(OEMHandler, self).get_health(summary)

    def get_licenses(self):
        if self.has_xcc:
            return self.immhandler.get_licenses()
        return super(OEMHandler, self).get_licenses()

    def get_user_expiration(self, uid):
        if self.has_xcc:
            return self.immhandler.get_user_expiration(uid)
        return None

    def delete_license(self, name):
        if self.has_xcc:
            return self.immhandler.delete_license(name)
        return super(OEMHandler, self).delete_license(name)

    def save_licenses(self, directory):
        if self.has_xcc:
            return self.immhandler.save_licenses(directory)
        return super(OEMHandler, self).save_licenses(directory)

    def apply_license(self, filename, progress=None, data=None):
        if self.has_xcc:
            return self.immhandler.apply_license(filename, progress, data)
        return super(OEMHandler, self).apply_license(filename, progress, data)

    def set_oem_extended_privilleges(self, uid):
        """Set user extended privillege as 'KVM & VMedia Allowed'

        |KVM & VMedia Not Allowed	0x00 0x00 0x00 0x00
        |KVM Only Allowed	0x01 0x00 0x00 0x00
        |VMedia Only Allowed	0x02  0x00 0x00 0x00
        |KVM & VMedia Allowed	0x03 0x00 0x00 0x00

        :param uid: User ID.
        """
        if self.has_tsm:
            self.ipmicmd.xraw_command(netfn=0x32, command=0xa3, data=(
                uid, 0x03, 0x00, 0x00, 0x00))
            return True
        return False

    def get_user_privilege_level(self, uid):
        if self.has_xcc:
            return self.immhandler.get_user_privilege_level(uid)
        return None

    def set_user_access(self, uid, channel, callback, link_auth, ipmi_msg, privilege_level):
        if self.has_xcc:
            self.immhandler.set_user_access(uid, privilege_level)

    def process_zero_fru(self, zerofru):
        if (self.oemid['manufacturer_id'] == 19046
                and self.oemid['product_id'] == 13616):
            # Currently SD350 FRU UUID is synchronized with the Device UUID.
            # Need to change to System UUID in future.
            # Since the IPMI get device uuid matches SMBIOS,
            # no need to decode it.
            guiddata = self.ipmicmd.raw_command(netfn=6, command=0x8)
            if 'error' not in guiddata:
                zerofru['UUID'] = util.decode_wireformat_uuid(
                    guiddata['data'], True)
        else:
            # It is expected that a manufacturer matches SMBIOS to IPMI
            # get system uuid return data.  If a manufacturer does not
            # do so, they should handle either deletion or fixup in the
            # OEM processing pass.  Code optimistically assumes that if
            # data is returned, than the vendor is properly using it.

            guiddata = self.ipmicmd.raw_command(netfn=6, command=0x37)
            if 'error' not in guiddata:
                if (self.oemid['manufacturer_id'] == 19046
                        and (self.oemid['product_id'] == 1182
                             or self.oemid['product_id'] == 1184)):
                    # The manufacturer (Asrockrack) of RS160/TS460
                    # matches SMBIOS
                    # to IPMI get system uuid return data,
                    # no need to decode it.
                    zerofru['UUID'] = util.decode_wireformat_uuid(
                        guiddata['data'], True)
                else:
                    zerofru['UUID'] = util.decode_wireformat_uuid(
                        guiddata['data'])
        if self.is_fpc:
            self.smmhandler.augment_zerofru(zerofru, self._fpc_variant)
        return self.process_fru(zerofru)

    def get_ami_sensor_reading(self, sensorname):
        """Get an OEM sensor

        If software wants to model some OEM behavior as a 'sensor' without
        doing SDR, this hook provides that ability.  It should mimic
        the behavior of 'get_sensor_reading' in command.py.
        """
        for sensor in self.get_sensor_data():
            if sensor.name == sensorname:
                return sensor

    def get_ami_sensor_descriptions(self):
        """Get list of OEM sensor names and types

        Iterate over dicts describing a label and type for OEM 'sensors'.  This
        should mimic the behavior of the get_sensor_descriptions function
        in command.py.
        """
        if self.has_ami:
            energy_sensor = energy.Energy(self.ipmicmd)
            for sensor in energy_sensor.get_energy_sensor():
                yield {'name': sensor.name,
                       'type': sensor.type}

    def get_ami_sensor_data(self):
        """Get OEM sensor data

        Iterate through all OEM 'sensors' and return data as if they were
        normal sensors.  This should mimic the behavior of the get_sensor_data
        function in command.py.
        """
        if self.has_ami:
            energy_sensor = energy.Energy(self.ipmicmd)
            for sensor in energy_sensor.get_energy_sensor():
                yield sensor



# coding=utf8
# Copyright 2014 IBM Corporation
# Copyright 2015 Lenovo
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module provides access to SDR offered by a BMC

This data is common between 'sensors' and 'inventory' modules since SDR
is both used to enumerate sensors for sensor commands and FRU ids for FRU
commands

For now, we will not offer persistent SDR caching as we do in xCAT's IPMI
code.  Will see if it is adequate to advocate for high object reuse in a
persistent process for the moment.

Focus is at least initially on the aspects that make the most sense for a
remote client to care about.  For example, smbus information is being
skipped for now
"""

import math
import os
import random
import string
import struct
import weakref

import six

import pyghmi.constants as const
import pyghmi.exceptions as exc


TYPE_UNKNOWN = 0
TYPE_SENSOR = 1
TYPE_FRU = 2

shared_sdrs = {}


oem_type_offsets = {
    343: {  # Intel
        149: {  # Cascade Lake-AP
            0x7a: {
                0xda: {
                    3: {
                        'desc': 'Allowed',
                        'severity': const.Health.Ok,
                    },
                    4: {
                        'desc': 'Restricted',
                        'severity': const.Health.Ok,
                    },
                    5: {
                        'desc': 'Disabled',
                        'severity': const.Health.Ok,
                    },
                },
            },
        },
    },
}


def ones_complement(value, bits):
    # utility function to help with the large amount of 2s
    # complement prevalent in ipmi spec
    signbit = 0b1 << (bits - 1)
    if value & signbit:
        # if negative, subtract 1, then take 1s
        # complement given bits width
        return 0 - (value ^ ((0b1 << bits) - 1))
    else:
        return value


def twos_complement(value, bits):
    # utility function to help with the large amount of 2s
    # complement prevalent in ipmi spec
    signbit = 0b1 << (bits - 1)
    if value & signbit:
        # if negative, subtract 1, then take 1s
        # complement given bits width
        return 0 - ((value - 1) ^ ((0b1 << bits) - 1))
    else:
        return value


unit_types = {
    # table 43-15 'sensor unit type codes'
    0: '',
    1: '°C',
    2: '°F',
    3: 'K',
    4: 'V',
    5: 'A',
    6: 'W',
    7: 'J',
    8: 'C',
    9: 'VA',
    10: 'nt',
    11: 'lm',
    12: 'lx',
    13: 'cd',
    14: 'kPa',
    15: 'PSI',
    16: 'N',
    17: 'CFM',
    18: 'RPM',
    19: 'Hz',
    20: 'μs',
    21: 'ms',
    22: 's',
    23: 'min',
    24: 'hr',
    25: 'd',
    26: 'week(s)',
    27: 'mil',
    28: 'inches',
    29: 'ft',
    30: 'cu in',
    31: 'cu feet',
    32: 'mm',
    33: 'cm',
    34: 'm',
    35: 'cu cm',
    36: 'cu m',
    37: 'L',
    38: 'fl. oz.',
    39: 'radians',
    40: 'steradians',
    41: 'revolutions',
    42: 'cycles',
    43: 'g',
    44: 'ounce',
    45: 'lb',
    46: 'ft-lb',
    47: 'oz-in',
    48: 'gauss',
    49: 'gilberts',
    50: 'henry',
    51: 'millihenry',
    52: 'farad',
    53: 'microfarad',
    54: 'ohms',
    55: 'siemens',
    56: 'mole',
    57: 'becquerel',
    58: 'ppm',
    60: 'dB',
    61: 'dBA',
    62: 'dBC',
    63: 'Gy',
    64: 'sievert',
    65: 'color temp deg K',
    66: 'bit',
    67: 'kb',
    68: 'mb',
    69: 'gb',
    70: 'byte',
    71: 'kB',
    72: 'mB',
    73: 'gB',
    74: 'word',
    75: 'dword',
    76: 'qword',
    77: 'line',
    78: 'hit',
    79: 'miss',
    80: 'retry',
    81: 'reset',
    82: 'overrun/overflow',
    83: 'underrun',
    84: 'collision',
    85: 'packets',
    86: 'messages',
    87: 'characters',
    88: 'error',
    89: 'uncorrectable error',
    90: 'correctable error',
    91: 'fatal error',
    92: 'grams',
}

sensor_rates = {
    0: '',
    1: ' per us',
    2: ' per ms',
    3: ' per s',
    4: ' per minute',
    5: ' per hour',
    6: ' per day',
}


class SensorReading(object):
    """Representation of the state of a sensor.

    It is initialized by pyghmi internally, it does not make sense for
    a developer to create one of these objects directly.

    It provides the following properties:
    name: UTF-8 string describing the sensor
    units: UTF-8 string describing the units of the sensor (if numeric)
    value: Value of the sensor if numeric
    imprecision: The amount by which the actual measured value may deviate from
        'value' due to limitations in the resolution of the given sensor.
    """

    def __init__(self, reading, suffix):
        self.broken_sensor_ids = {}
        self.health = const.Health.Ok
        self.type = reading['type']
        self.value = None
        self.imprecision = None
        self.states = []
        self.state_ids = []
        self.unavailable = 0
        try:
            self.health = reading['health']
            self.states = reading['states']
            self.state_ids = reading['state_ids']
            self.value = reading['value']
            self.imprecision = reading['imprecision']
        except KeyError:
            pass
        if 'unavailable' in reading:
            self.unavailable = 1
        self.units = suffix
        self.name = reading['name']

    def __repr__(self):
        return repr({
            'value': self.value,
            'states': self.states,
            'state_ids': self.state_ids,
            'units': self.units,
            'imprecision': self.imprecision,
            'name': self.name,
            'type': self.type,
            'unavailable': self.unavailable,
            'health': self.health
        })

    def simplestring(self):
        """Return a summary string of the reading.

        This is intended as a sampling of how the data could be presented by
        a UI.  It's intended to help a developer understand the relation
        between the attributes of a sensor reading if it is not quite clear
        """
        repr = self.name + ": "
        if self.value is not None:
            repr += str(self.value)
            repr += " ± " + str(self.imprecision)
            repr += self.units
        for state in self.states:
            repr += state + ","
        if self.health >= const.Health.Failed:
            repr += '(Failed)'
        elif self.health >= const.Health.Critical:
            repr += '(Critical)'
        elif self.health >= const.Health.Warning:
            repr += '(Warning)'
        return repr


class SDREntry(object):
    """Represent a single entry in the IPMI SDR.

    This is created and consumed by pyghmi internally, there is no reason for
    external code to pay attention to this class.
    """

    def __init__(self, entrybytes, event_consts, reportunsupported=False,
                 mfg_id=0, prod_id=0):
        self.mfg_id = mfg_id
        self.prod_id = prod_id
        self.event_consts = event_consts
        # ignore record id for now, we only care about the sensor number for
        # moment
        self.readable = True
        self.reportunsupported = reportunsupported
        if entrybytes[2] != 0x51:
            # only recognize '1.5', the only version defined at time of writing
            raise NotImplementedError
        self.rectype = entrybytes[3]
        self.linearization = None
        # most important to get going are 1, 2, and 11
        self.sdrtype = TYPE_SENSOR  # assume a sensor
        if self.rectype == 1:  # full sdr
            self.full_decode(entrybytes[5:])
        elif self.rectype == 2:  # full sdr
            self.compact_decode(entrybytes[5:])
        elif self.rectype == 3:  # event only
            self.eventonly_decode(entrybytes[5:])
        elif self.rectype == 8:  # entity association
            self.association_decode(entrybytes[5:])
        elif self.rectype == 0x11:  # FRU locator
            self.fru_decode(entrybytes[5:])
        elif self.rectype == 0x12:  # Management controller
            self.mclocate_decode(entrybytes[5:])
        elif self.rectype == 0xc0:  # OEM format
            self.sdrtype = TYPE_UNKNOWN   # assume undefined
            self.oem_decode(entrybytes[5:])
        elif self.reportunsupported:
            raise NotImplementedError
        else:
            self.sdrtype = TYPE_UNKNOWN

    @property
    def name(self):
        if self.sdrtype == TYPE_SENSOR:
            return self.sensor_name
        elif self.sdrtype == TYPE_FRU:
            return self.fru_name
        else:
            return "UNKNOWN"

    def oem_decode(self, entry):
        mfgid = entry[0] + (entry[1] << 8) + (entry[2] << 16)
        if self.reportunsupported:
            raise NotImplementedError("No support for mfgid %X" % mfgid)

    def mclocate_decode(self, entry):
        # For now, we don't have use for MC locator records
        # we'll ignore them at the moment
        self.sdrtype = TYPE_UNKNOWN
        pass

    def fru_decode(self, entry):
        # table 43-7 FRU Device Locator
        self.sdrtype = TYPE_FRU
        self.fru_name = self.tlv_decode(entry[10], entry[11:])
        self.fru_number = entry[1]
        self.fru_logical = (entry[2] & 0b10000000) == 0b10000000
        # 0x8  to 0x10..  0 unspecified except on 0x10, 1 is dimm
        self.fru_type_and_modifier = (entry[5] << 8) + entry[6]

    def association_decode(self, entry):
        # table 43-4 Entity Associaition Record
        # TODO(jbjohnso): actually represent this data
        self.sdrtype = TYPE_UNKNOWN

    def eventonly_decode(self, entry):
        # table 43-3 event_only sensor record
        self._common_decode(entry)
        self.sensor_name = self.tlv_decode(entry[11], entry[12:])
        self.readable = False

    def compact_decode(self, entry):
        # table 43-2 compact sensor record
        self._common_decode(entry)
        self.sensor_name = self.tlv_decode(entry[26], entry[27:])

    def assert_trap_value(self, offset):
        trapval = (self.sensor_type_number << 16) + (self.reading_type << 8)
        return trapval + offset

    def _common_decode(self, entry):
        # event only, compact and full are very similar
        # this function handles the common aspects of compact and full
        # offsets from spec, minus 6
        self.sensor_owner = entry[0]
        self.sensor_lun = entry[1] & 0x03
        self.sensor_number = entry[2]
        self.entity = self.event_consts.entity_ids.get(
            entry[3], 'Unknown entity {0}'.format(entry[3]))
        if self.rectype == 3:
            self.sensor_type_number = entry[5]
            self.reading_type = entry[6]  # table 42-1
        else:
            self.sensor_type_number = entry[7]
            self.reading_type = entry[8]  # table 42-1
        try:
            self.sensor_type = self.event_consts.sensor_type_codes[
                self.sensor_type_number]
        except KeyError:
            self.sensor_type = "UNKNOWN type " + str(self.sensor_type_number)
        if self.rectype == 3:
            return
        # 0: unspecified
        # 1: generic threshold based
        # 0x6f: discrete sensor-specific from table 42-3, sensor offsets
        # all others per table 42-2, generic discrete
        # numeric format is one of:
        # 0 - unsigned, 1 - 1s complement, 2 - 2s complement, 3 - ignore number
        # compact records are supposed to always write it as '3', presumably
        # to allow for the concept of a compact record with a numeric format
        # even though numerics are not allowed today.  Some implementations
        # violate the spec and do something other than 3 today.  Tolerate
        # the violation under the assumption that things are not so hard up
        # that there will ever be a need for compact sensors supporting numeric
        # values
        if self.rectype == 2:
            self.numeric_format = 3
        else:
            self.numeric_format = (entry[15] & 0b11000000) >> 6
        self.sensor_rate = sensor_rates[(entry[15] & 0b111000) >> 3]
        self.unit_mod = ""
        if (entry[15] & 0b110) == 0b10:  # unit1 by unit2
            self.unit_mod = "/"
        elif (entry[15] & 0b110) == 0b100:
            # combine the units by multiplying, SI nomenclature is either spac
            # or hyphen, so go with space
            self.unit_mod = " "
        self.percent = ''
        if entry[15] & 1 == 1:
            self.percent = '% '
        if self.sensor_type_number == 0xb:
            if self.unit_mod == '':
                if entry[16] == 6:
                    self.sensor_type = 'Power'
            elif self.unit_mod == ' ':
                if entry[16] == 6 and entry[17] in (22, 23, 24):
                    self.sensor_type = 'Energy'
        self.baseunit = unit_types[entry[16]]
        self.modunit = unit_types[entry[17]]
        self.unit_suffix = self.percent + self.baseunit + self.unit_mod + \
            self.modunit

    def full_decode(self, entry):
        # offsets are table from spec, minus 6
        # TODO(jbjohnso): table 43-13, put in constants to interpret entry[3]
        self._common_decode(entry)
        # now must extract the formula data to transform values
        # entry[18 to entry[24].
        # if not linear, must use get sensor reading factors
        # TODO(jbjohnso): the various other values
        self.sensor_name = self.tlv_decode(entry[42], entry[43:])
        self.linearization = entry[18] & 0b1111111
        if self.linearization <= 11:
            # the enumuration of linear sensors goes to 11,
            # static formula parameters are applicable, decode them
            # if 0x70, then the sesor reading will have to get the
            # factors on the fly.
            # the formula could apply if we bother with nominal
            # reading interpretation
            self.decode_formula(entry[19:25])

    def _decode_state(self, state):
        mapping = self.event_consts.generic_type_offsets
        try:
            if self.reading_type in mapping:
                desc = mapping[self.reading_type][state]['desc']
                health = mapping[self.reading_type][state]['severity']
            elif self.reading_type == 0x6f:
                mapping = self.event_consts.sensor_type_offsets
                desc = mapping[self.sensor_type_number][state]['desc']
                health = mapping[self.sensor_type_number][state]['severity']
            elif self.reading_type >= 0x70 and self.reading_type <= 0x7f:
                sensedata = oem_type_offsets[self.mfg_id][self.prod_id][
                    self.reading_type][self.sensor_type_number][state]
                desc = sensedata['desc']
                health = sensedata['severity']
            else:
                desc = "Unknown state %d" % state
                health = const.Health.Warning
        except KeyError:
            desc = "Unknown state %d for reading type %d/sensor type %d" % (
                state, self.reading_type, self.sensor_type_number)
            health = const.Health.Warning
        return desc, health

    def decode_sensor_reading(self, ipmicmd, reading):
        numeric = None
        output = {
            'name': self.sensor_name,
            'type': self.sensor_type,
            'id': self.sensor_number,
        }
        if reading[1] & 0b100000 or not reading[1] & 0b1000000:
            output['unavailable'] = 1
            return SensorReading(output, self.unit_suffix)
        if self.numeric_format == 2:
            numeric = twos_complement(reading[0], 8)
        elif self.numeric_format == 1:
            numeric = ones_complement(reading[0], 8)
        elif self.numeric_format == 0:
            numeric = reading[0]
        discrete = True
        if numeric is not None:
            lowerbound = numeric - (0.5 + (self.tolerance / 2.0))
            upperbound = numeric + (0.5 + (self.tolerance / 2.0))
            lowerbound = self.decode_value(ipmicmd, lowerbound)
            upperbound = self.decode_value(ipmicmd, upperbound)
            output['value'] = (lowerbound + upperbound) / 2.0
            output['imprecision'] = output['value'] - lowerbound
            discrete = False
        upper = 'upper'
        lower = 'lower'
        if self.linearization == 7:
            # if the formula is 1/x, then the intuitive sense of upper and
            # lower are backwards
            upper = 'lower'
            lower = 'upper'
        output['states'] = []
        output['state_ids'] = []
        output['health'] = const.Health.Ok
        if discrete:
            for state in range(8):
                if reading[2] & (0b1 << state):
                    statedesc, health = self._decode_state(state)
                    output['health'] |= health
                    output['states'].append(statedesc)
                    output['state_ids'].append(self.assert_trap_value(state))
            if len(reading) > 3:
                for state in range(7):
                    if reading[3] & (0b1 << state):
                        statedesc, health = self._decode_state(state + 8)
                        output['health'] |= health
                        output['states'].append(statedesc)
                        output['state_ids'].append(
                            self.assert_trap_value(state + 8))
        else:
            if reading[2] & 0b1:
                output['health'] |= const.Health.Warning
                output['states'].append(lower + " non-critical threshold")
                output['state_ids'].append(self.assert_trap_value(1))
            if reading[2] & 0b10:
                output['health'] |= const.Health.Critical
                output['states'].append(lower + " critical threshold")
                output['state_ids'].append(self.assert_trap_value(2))
            if reading[2] & 0b100:
                output['health'] |= const.Health.Failed
                output['states'].append(lower + " non-recoverable threshold")
                output['state_ids'].append(self.assert_trap_value(3))
            if reading[2] & 0b1000:
                output['health'] |= const.Health.Warning
                output['states'].append(upper + " non-critical threshold")
                output['state_ids'].append(self.assert_trap_value(4))
            if reading[2] & 0b10000:
                output['health'] |= const.Health.Critical
                output['states'].append(upper + " critical threshold")
                output['state_ids'].append(self.assert_trap_value(5))
            if reading[2] & 0b100000:
                output['health'] |= const.Health.Failed
                output['states'].append(upper + " non-recoverable threshold")
                output['state_ids'].append(self.assert_trap_value(6))
        return SensorReading(output, self.unit_suffix)

    def _set_tmp_formula(self, ipmicmd, value):
        rsp = ipmicmd.raw_command(netfn=4, command=0x23,
                                  data=(self.sensor_number, value))
        # skip next reading field, not used in on-demand situation
        self.decode_formula(rsp['data'][1:])

    def decode_value(self, ipmicmd, value):
        # Take the input value and return meaningful value
        linearization = self.linearization
        if linearization > 11:  # direct calling code to get factors
            # for now, we will get the factors on demand
            # the facility is engineered such that at construction
            # time the entire BMC table should be fetchable in a reasonable
            # fashion.  However for now opt for retrieving rows as needed
            # rather than tracking all that information for a relatively
            # rare behavior
            self._set_tmp_formula(ipmicmd, value)
            linearization = 0
        # time to compute the pre-linearization value.
        decoded = float((value * self.m + self.b)
                        * (10 ** self.resultexponent))
        if linearization == 0:
            return decoded
        elif linearization == 1:
            return math.log(decoded)
        elif linearization == 2:
            return math.log(decoded, 10)
        elif linearization == 3:
            return math.log(decoded, 2)
        elif linearization == 4:
            return math.exp(decoded)
        elif linearization == 5:
            return 10 ** decoded
        elif linearization == 6:
            return 2 ** decoded
        elif linearization == 7:
            return 1 / decoded
        elif linearization == 8:
            return decoded ** 2
        elif linearization == 9:
            return decoded ** 3
        elif linearization == 10:
            return math.sqrt(decoded)
        elif linearization == 11:
            return decoded ** (1.0 / 3)
        else:
            raise NotImplementedError

    def decode_formula(self, entry):
        self.m = twos_complement(entry[0] + ((entry[1] & 0b11000000) << 2), 10)
        self.tolerance = entry[1] & 0b111111
        self.b = twos_complement(entry[2] + ((entry[3] & 0b11000000) << 2), 10)
        self.accuracy = (entry[3] & 0b111111) + (entry[4] & 0b11110000) << 2
        self.accuracyexp = (entry[4] & 0b1100) >> 2
        self.direction = entry[4] & 0b11
        # 0 = n/a, 1 = input, 2 = output
        self.resultexponent = twos_complement((entry[5] & 0b11110000) >> 4, 4)
        bexponent = twos_complement(entry[5] & 0b1111, 4)
        # might as well do the math to 'b' now rather than wait for later
        self.b = self.b * (10**bexponent)

    def tlv_decode(self, tlv, data):
        # Per IPMI 'type/length byte format
        ipmitype = (tlv & 0b11000000) >> 6
        if not len(data):
            return ""
        if ipmitype == 0:  # Unicode per 43.15 in ipmi 2.0 spec
            # the spec is not specific about encoding, assuming utf8
            return six.text_type(struct.pack("%dB" % len(data), *data),
                                 "utf_8")
        elif ipmitype == 1:  # BCD '+'
            tmpl = "%02X" * len(data)
            tstr = tmpl % tuple(data)
            tstr = tstr.replace("A", " ").replace("B", "-").replace("C", ".")
            return tstr.replace("D", ":").replace("E", ",").replace("F", "_")
        elif ipmitype == 2:  # 6 bit ascii, start at 0x20
            # the ordering is very peculiar and is best understood from
            # IPMI SPEC "6-bit packed ascii example
            tstr = ""
            while len(data) >= 3:  # the packing only works with 3 byte chunks
                tstr += chr((data[0] & 0b111111) + 0x20)
                tstr += chr(((data[1] & 0b1111) << 2) + (data[0] >> 6) + 0x20)
                tstr += chr(((data[2] & 0b11) << 4) + (data[1] >> 4) + 0x20)
                tstr += chr((data[2] >> 2) + 0x20)
            if not isinstance(tstr, str):
                tstr = tstr.decode('utf-8')
            return tstr
        elif ipmitype == 3:  # ACSII+LATIN1
            ret = struct.pack("%dB" % len(data), *data)
            if not isinstance(ret, str):
                ret = ret.decode('utf-8')
            return ret


class SDR(object):
    """Examine the state of sensors managed by a BMC

    Presents the data from sensor read commands as directed by the SDR in a
    reasonable format.  This module is used by the command module, and is not
    intended for consumption by external code directly

    :param ipmicmd: A Command class object
    """
    def __init__(self, ipmicmd, cachedir=None):
        self.ipmicmd = weakref.proxy(ipmicmd)
        self.sensors = {}
        self.fru = {}
        self.cachedir = cachedir
        self.read_info()

    def read_info(self):
        # first, we want to know the device id
        rsp = self.ipmicmd.xraw_command(netfn=6, command=1)
        rsp['data'] = bytearray(rsp['data'])
        self.device_id = rsp['data'][0]
        self.device_rev = rsp['data'][1] & 0b111
        # Going to ignore device available until get sdr command
        # since that provides usefully distinct state and this does not
        self.fw_major = rsp['data'][2] & 0b1111111
        self.fw_minor = "%02X" % rsp['data'][3]  # BCD encoding, oddly enough
        self.ipmiversion = rsp['data'][4]  # 51h = 1.5, 02h = 2.0
        self.mfg_id = (rsp['data'][8] << 16) + (rsp['data'][7] << 8) + \
            rsp['data'][6]
        self.prod_id = (rsp['data'][10] << 8) + rsp['data'][9]
        if len(rsp['data']) > 11:
            self.aux_fw = self.decode_aux(rsp['data'][11:15])
        if rsp['data'][1] & 0b10000000 and rsp['data'][5] & 0b10 == 0:
            # The device has device sdrs, also does not support SDR repository
            # device, so we are meant to use an alternative mechanism to get
            # SDR data
            if rsp['data'][5] & 1:
                # The device has sensor device support, so in theory we should
                # be able to proceed
                # However at the moment, we haven't done so
                raise NotImplementedError
            return
            # We have Device SDR, without SDR Repository device, but
            # also without sensor device support, no idea how to
            # continue
        self.get_sdr()

    def get_sdr_reservation(self):
        rsp = self.ipmicmd.raw_command(netfn=0x0a, command=0x22)
        if rsp['code'] != 0:
            raise exc.IpmiException(rsp['error'])
        return rsp['data'][0] + (rsp['data'][1] << 8)

    def get_sdr(self):
        repinfo = self.ipmicmd.xraw_command(netfn=0x0a, command=0x20)
        repinfo['data'] = bytearray(repinfo['data'])
        if (repinfo['data'][0] != 0x51):
            # we only understand SDR version 51h, the only version defined
            # at time of this writing
            raise NotImplementedError
        # NOTE(jbjohnso): we actually don't need to care about 'numrecords'
        # since FFFF marks the end explicitly
        # numrecords = (rsp['data'][2] << 8) + rsp['data'][1]
        # NOTE(jbjohnso): don't care about 'free space' at the moment
        # NOTE(jbjohnso): most recent timstamp data for add and erase could be
        # handy to detect cache staleness, but for now will assume invariant
        # over life of session
        # NOTE(jbjohnso): not looking to support the various options in op
        # support, ignore those for now, reservation if some BMCs can't read
        # full SDR in one slurp
        modtime = struct.unpack('!Q', bytes(repinfo['data'][5:13]))[0]
        recid = 0
        rsvid = 0  # partial 'get sdr' will require this
        offset = 0
        size = 0xff
        chunksize = 128
        try:
            csdrs = shared_sdrs[
                (self.fw_major, self.fw_minor, self.mfg_id, self.prod_id,
                 self.device_id, modtime)]
            self.sensors = csdrs['sensors']
            self.fru = csdrs['fru']
            return
        except KeyError:
            pass
        cachefilename = None
        self.broken_sensor_ids = {}
        if self.cachedir:
            cachefilename = 'sdrcache-2.{0}.{1}.{2}.{3}.{4}.{5}'.format(
                self.mfg_id, self.prod_id, self.device_id, self.fw_major,
                self.fw_minor, modtime)
            cachefilename = os.path.join(self.cachedir, cachefilename)
        if cachefilename and os.path.isfile(cachefilename):
            with open(cachefilename, 'rb') as cfile:
                csdrlen = cfile.read(2)
                while csdrlen:
                    csdrlen = struct.unpack('!H', csdrlen)[0]
                    self.add_sdr(cfile.read(csdrlen))
                    csdrlen = cfile.read(2)
                for sid in self.broken_sensor_ids:
                    try:
                        del self.sensors[sid]
                    except KeyError:
                        pass
                shared_sdrs[
                    (self.fw_major, self.fw_minor, self.mfg_id, self.prod_id,
                     self.device_id, modtime)] = {
                    'sensors': self.sensors,
                    'fru': self.fru,
                }
                return
        sdrraw = [] if cachefilename else None
        while recid != 0xffff:  # per 33.12 Get SDR command, 0xffff marks end
            newrecid = 0
            currlen = 0
            sdrdata = bytearray()
            while True:  # loop until SDR fetched wholly
                if size != 0xff and rsvid == 0:
                    rsvid = self.get_sdr_reservation()
                rqdata = [rsvid & 0xff, rsvid >> 8,
                          recid & 0xff, recid >> 8,
                          offset, size]
                sdrrec = self.ipmicmd.raw_command(netfn=0x0a, command=0x23,
                                                  data=rqdata)
                if sdrrec['code'] == 0xca:
                    if size == 0xff:  # get just 5 to get header to know length
                        size = 5
                    elif size > 5:
                        size //= 2
                        # push things over such that it's less
                        # likely to be just 1 short of a read
                        # and incur a whole new request
                        size += 2
                        chunksize = size
                    continue
                if sdrrec['code'] == 0xc5:  # need a new reservation id
                    rsvid = 0
                    continue
                if sdrrec['code'] != 0:
                    raise exc.IpmiException(sdrrec['error'])
                if newrecid == 0:
                    newrecid = (sdrrec['data'][1] << 8) + sdrrec['data'][0]
                if currlen == 0:
                    currlen = sdrrec['data'][6] + 5  # compensate for header
                sdrdata.extend(sdrrec['data'][2:])
                # determine next offset to use based on current offset and the
                # size used last time.
                offset += size
                if offset >= currlen:
                    break
                if size == 5 and offset == 5:
                    # bump up size after header retrieval
                    size = chunksize
                if (offset + size) > currlen:
                    size = currlen - offset
            self.add_sdr(sdrdata)
            if sdrraw is not None:
                sdrraw.append(bytes(sdrdata))
            offset = 0
            if size != 0xff:
                size = 5
            if newrecid == recid:
                raise exc.BmcErrorException("Incorrect SDR record id from BMC")
            recid = newrecid
        for sid in self.broken_sensor_ids:
            try:
                del self.sensors[sid]
            except KeyError:
                pass
        shared_sdrs[(self.fw_major, self.fw_minor, self.mfg_id, self.prod_id,
                     self.device_id, modtime)] = {
            'sensors': self.sensors,
            'fru': self.fru,
        }
        if cachefilename:
            suffix = ''.join(
                random.choice(string.ascii_lowercase) for _ in range(12))
            with open(cachefilename + '.' + suffix, 'wb') as cfile:
                for csdr in sdrraw:
                    cfile.write(struct.pack('!H', len(csdr)))
                    cfile.write(csdr)
            os.rename(cachefilename + '.' + suffix, cachefilename)

    def get_sensor_numbers(self):
        for number in self.sensors:
            if self.sensors[number].readable:
                yield number

    def make_sdr_entry(self, sdrbytes):
        return SDREntry(sdrbytes, self.ipmicmd.get_event_constants(),
                        False, self.mfg_id, self.prod_id)

    def add_sdr(self, sdrbytes):
        if not isinstance(sdrbytes[0], int):
            sdrbytes = bytearray(sdrbytes)
        newent = self.make_sdr_entry(sdrbytes)
        if newent.sdrtype == TYPE_SENSOR:
            id = '{0}.{1}.{2}'.format(
                newent.sensor_owner, newent.sensor_number, newent.sensor_lun)
            if id in self.sensors:
                self.broken_sensor_ids[id] = True
                return
            self.sensors[id] = newent
        elif newent.sdrtype == TYPE_FRU:
            id = newent.fru_number
            if id in self.fru:
                self.broken_sensor_ids[id] = True
                return
            self.fru[id] = newent

    def decode_aux(self, auxdata):
        # This is where manufacturers can add their own
        # decode information
        return "".join(hex(x) for x in auxdata)



# Copyright 2015 Lenovo
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyghmi.ipmi.oem.lenovo import inventory


dimm_fields = (
    inventory.EntryField("index", "B"),
    inventory.EntryField("manufacture_location", "B"),
    inventory.EntryField("channel_number", "B"),
    inventory.EntryField("module_type", "10s"),
    inventory.EntryField("ddr_voltage", "10s"),
    inventory.EntryField("speed", "<h",
                         valuefunc=lambda v: str(v) + " MHz"),
    inventory.EntryField("capacity_mb", "<h",
                         valuefunc=lambda v: v * 1024),
    inventory.EntryField("manufacturer", "30s"),
    inventory.EntryField("serial", ">I",
                         valuefunc=lambda v: hex(v)[2:]),
    inventory.EntryField("model", "21s"),
    inventory.EntryField("reserved", "h", include=False)
)

dimm_cmd = {
    "lenovo": {
        "netfn": 0x06,
        "command": 0x59,
        "data": (0x00, 0xc1, 0x02, 0x00)},
    "asrock": {
        "netfn": 0x3a,
        "command": 0x50,
        "data": (0x01, 0x02, 0x01)},
}


def parse_dimm_info(raw):
    return inventory.parse_inventory_category_entry(raw, dimm_fields)


def get_categories():
    return {
        "dimm": {
            "idstr": "DIMM {0}",
            "parser": parse_dimm_info,
            "command": dimm_cmd,
            "workaround_bmc_bug": lambda t: True
        }
    }



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This is a simple, but working proof of concept of using pyghmi.ipmi.bmc to
control a VM
"""

import argparse
import sys
import threading

import libvirt

import pyghmi.ipmi.bmc as bmc


def lifecycle_callback(connection, domain, event, detail, console):
    console.state = console.domain.state(0)


def error_handler(unused, error):
    if (error[0] == libvirt.VIR_ERR_RPC
            and error[1] == libvirt.VIR_FROM_STREAMS):
        return


def stream_callback(stream, events, console):
    try:
        data = console.stream.recv(1024)
    except Exception:
        return
    if console.sol:
        console.sol.send_data(data)


class LibvirtBmc(bmc.Bmc):
    """A class to provide an IPMI interface to the VirtualBox APIs."""

    def __init__(self, authdata, hypervisor, domain, port):
        super(LibvirtBmc, self).__init__(authdata, port)
        # Rely on libvirt to throw on bad data
        self.conn = libvirt.open(hypervisor)
        self.name = domain
        self.domain = self.conn.lookupByName(domain)
        self.state = self.domain.state(0)
        self.stream = None
        self.run_console = False
        self.conn.domainEventRegister(lifecycle_callback, self)
        self.sol_thread = None

    def cold_reset(self):
        # Reset of the BMC, not managed system, here we will exit the demo
        print('shutting down in response to BMC cold reset request')
        sys.exit(0)

    def get_power_state(self):
        if self.domain.isActive():
            return 'on'
        else:
            return 'off'

    def power_off(self):
        if not self.domain.isActive():
            return 0xd5  # Not valid in this state
        self.domain.destroy()

    def power_on(self):
        if self.domain.isActive():
            return 0xd5  # Not valid in this state
        self.domain.create()

    def power_reset(self):
        if not self.domain.isActive():
            return 0xd5  # Not valid in this state
        self.domain.reset()

    def power_shutdown(self):
        if not self.domain.isActive():
            return 0xd5  # Not valid in this state
        self.domain.shutdown()

    def is_active(self):
        return self.domain.isActive()

    def check_console(self):
        if (self.state[0] == libvirt.VIR_DOMAIN_RUNNING
                or self.state[0] == libvirt.VIR_DOMAIN_PAUSED):
            if self.stream is None:
                self.stream = self.conn.newStream(libvirt.VIR_STREAM_NONBLOCK)
                self.domain.openConsole(None, self.stream, 0)
                self.stream.eventAddCallback(libvirt.VIR_STREAM_EVENT_READABLE,
                                             stream_callback, self)
        else:
            if self.stream:
                self.stream.eventRemoveCallback()
                self.stream = None

        return self.run_console

    def activate_payload(self, request, session):
        super(LibvirtBmc, self).activate_payload(request, session)
        self.run_console = True
        self.sol_thread = threading.Thread(target=self.loop)
        self.sol_thread.start()

    def deactivate_payload(self, request, session):
        self.run_console = False
        self.sol_thread.join()
        super(LibvirtBmc, self).deactivate_payload(request, session)

    def iohandler(self, data):
        if self.stream:
            self.stream.send(data)

    def loop(self):
        while self.check_console():
            libvirt.virEventRunDefaultImpl()


def main():
    parser = argparse.ArgumentParser(
        prog='virshbmc',
        description='Pretend to be a BMC and proxy to virsh',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument('--port',
                        dest='port',
                        type=int,
                        default=623,
                        help='(UDP) port to listen on')
    parser.add_argument('--connect',
                        dest='hypervisor',
                        default='qemu:///system',
                        help='The hypervisor to connect to')
    parser.add_argument('--domain',
                        dest='domain',
                        required=True,
                        help='The name of the domain to manage')
    args = parser.parse_args()

    libvirt.virEventRegisterDefaultImpl()
    libvirt.registerErrorHandler(error_handler, None)

    mybmc = LibvirtBmc({'admin': 'password'},
                       hypervisor=args.hypervisor,
                       domain=args.domain,
                       port=args.port)
    mybmc.listen()


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2014 IBM Corporation
# Copyright 2015-2019 Lenovo
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This represents the low layer message framing portion of IPMI"""

import struct
import threading

import pyghmi.exceptions as exc
from pyghmi.ipmi.private import constants
from pyghmi.ipmi.private import session
from pyghmi.ipmi.private.util import _monotonic_time


class Console(object):
    """IPMI SOL class.

    This object represents an SOL channel, multiplexing SOL data with
    commands issued by ipmi.command.

    :param bmc: hostname or ip address of BMC
    :param userid: username to use to connect
    :param password: password to connect to the BMC
    :param iohandler: Either a function to call with bytes, a filehandle to
                      use for input and output, or a tuple of (input, output)
                      handles
    :param force: Set to True to force on or False to force off
    :param kg: optional parameter for BMCs configured to require it
    """

    # TODO(jbjohnso): still need an exit and a data callin function
    def __init__(self, bmc, userid, password,
                 iohandler, port=623,
                 force=False, kg=None):
        self.outputlock = threading.RLock()
        self.keepaliveid = None
        self.connected = False
        self.broken = False
        self.out_handler = iohandler
        self.remseq = 0
        self.myseq = 0
        self.lastsize = 0
        self.retriedpayload = 0
        self.pendingoutput = []
        self.awaitingack = False
        self.activated = False
        self.force_session = force
        self.port = port
        self.ipmi_session = None
        self.callgotsession = None
        self.ipmi_session = session.Session(bmc=bmc,
                                            userid=userid,
                                            password=password,
                                            port=port,
                                            kg=kg,
                                            onlogon=self._got_session)
        # induce one iteration of the loop, now that we would be
        # prepared for it in theory
        session.Session.wait_for_rsp(0)
        if self.callgotsession is not None:
            self._got_session(self.callgotsession)
            self.callgotsession = None

    def _got_session(self, response):
        """Private function to navigate SOL payload activation"""

        if 'error' in response:
            self._print_error(response['error'])
            return
        if not self.ipmi_session:
            self.callgotsession = response
            return
        # Send activate sol payload directive
        # netfn= 6 (application)
        # command = 0x48 (activate payload)
        # data = (1, sol payload type
        #        1, first instance
        #        0b11000000, -encrypt, authenticate,
        #                      disable serial/modem alerts, CTS fine
        #        0, 0, 0 reserved
        response = self.ipmi_session.raw_command(netfn=0x6, command=0x48,
                                                 data=(1, 1, 192, 0, 0, 0))
        # given that these are specific to the command,
        # it's probably best if one can grep the error
        # here instead of in constants
        sol_activate_codes = {
            0x81: 'SOL is disabled',
            0x82: 'Maximum SOL session count reached',
            0x83: 'Cannot activate payload with encryption',
            0x84: 'Cannot activate payload without encryption',
        }
        if 'code' in response and response['code']:
            if response['code'] in constants.ipmi_completion_codes:
                self._print_error(
                    constants.ipmi_completion_codes[response['code']])
                return
            elif response['code'] == 0x80:
                if self.force_session and not self.retriedpayload:
                    self.retriedpayload = 1
                    sessrsp = self.ipmi_session.raw_command(
                        netfn=0x6,
                        command=0x49,
                        data=(1, 1, 0, 0, 0, 0))
                    self._got_session(sessrsp)
                    return
                else:
                    self._print_error('SOL Session active for another client')
                    return
            elif response['code'] in sol_activate_codes:
                self._print_error(sol_activate_codes[response['code']])
                return
            else:
                self._print_error(
                    'SOL encountered Unrecognized error code %d' %
                    response['code'])
                return
        if 'error' in response:
            self._print_error(response['error'])
            return
        self.activated = True
        # data[0:3] is reserved except for the test mode, which we don't use
        data = response['data']
        self.maxoutcount = (data[5] << 8) + data[4]
        # BMC tells us this is the maximum allowed size
        # data[6:7] is the promise of how small packets are going to be, but we
        # don't have any reason to worry about it
        # some BMCs disagree on the endianness, so do both
        valid_ports = (self.port, struct.unpack(
            '<H', struct.pack('>H', self.port))[0])
        if (data[8] + (data[9] << 8)) not in valid_ports:
            # TODO(jbjohnso): support atypical SOL port number
            raise NotImplementedError("Non-standard SOL Port Number")
        # ignore data[10:11] for now, the vlan detail, shouldn't matter to this
        # code anyway...
        # NOTE(jbjohnso):
        # We will use a special purpose keepalive
        if self.ipmi_session.sol_handler is not None:
            # If there is erroneously another SOL handler already, notify
            # it of newly established session
            self.ipmi_session.sol_handler({'error': 'Session Disconnected'})
        self.keepaliveid = self.ipmi_session.register_keepalive(
            cmd={'netfn': 6, 'command': 0x4b, 'data': (1, 1)},
            callback=self._got_payload_instance_info)
        self.ipmi_session.sol_handler = self._got_sol_payload
        self.connected = True
        # self._sendpendingoutput() checks len(self._sendpendingoutput)
        self._sendpendingoutput()

    def _got_payload_instance_info(self, response):
        if 'error' in response:
            self.activated = False
            self._print_error(response['error'])
            return
        currowner = struct.unpack(
            "<I", struct.pack('4B', *response['data'][:4]))
        if currowner[0] != self.ipmi_session.sessionid:
            # the session is deactivated or active for something else
            self.activated = False
            self._print_error('SOL deactivated')
            return
        # ok, still here, that means session is alive, but another
        # common issue is firmware messing with mux on reboot
        # this would be a nice thing to check, but the serial channel
        # number is needed and there isn't an obvious means to reliably
        # discern which channel or even *if* the serial port in question
        # correlates at all to an ipmi channel to check mux

    def _addpendingdata(self, data):
        with self.outputlock:
            if isinstance(data, dict):
                self.pendingoutput.append(data)
            else:  # it is a text situation
                if (len(self.pendingoutput) == 0
                        or isinstance(self.pendingoutput[-1], dict)):
                    self.pendingoutput.append(data)
                else:
                    self.pendingoutput[-1] += data

    def _got_cons_input(self, handle):
        """Callback for handle events detected by ipmi session"""

        self._addpendingdata(handle.read())
        if not self.awaitingack:
            self._sendpendingoutput()

    def close(self):
        """Shut down an SOL session"""

        if self.ipmi_session:
            self.ipmi_session.unregister_keepalive(self.keepaliveid)
        if self.activated and self.ipmi_session is not None:
            try:
                self.ipmi_session.raw_command(netfn=6, command=0x49,
                                              data=(1, 1, 0, 0, 0, 0))
            except exc.IpmiException:
                # if underlying ipmi session is not working, then
                # run with the implicit success
                pass

    def send_data(self, data):
        if self.broken:
            return
        self._addpendingdata(data)
        if not self.connected:
            return
        if not self.awaitingack:
            self._sendpendingoutput()

    def send_break(self):
        self._addpendingdata({'break': 1})
        if not self.connected:
            return
        if not self.awaitingack:
            self._sendpendingoutput()

    @classmethod
    def wait_for_rsp(cls, timeout):
        """Delay for no longer than timeout for next response.

        This acts like a sleep that exits on activity.

        :param timeout: Maximum number of seconds before returning
        """
        return session.Session.wait_for_rsp(timeout=timeout)

    def _sendpendingoutput(self):
        with self.outputlock:
            dobreak = False
            chunk = ''
            if len(self.pendingoutput) == 0:
                return
            if isinstance(self.pendingoutput[0], dict):
                if 'break' in self.pendingoutput[0]:
                    dobreak = True
                else:
                    del self.pendingoutput[0]
                    raise ValueError
                del self.pendingoutput[0]
            elif len(self.pendingoutput[0]) > self.maxoutcount:
                chunk = self.pendingoutput[0][:self.maxoutcount]
                self.pendingoutput[0] = self.pendingoutput[0][
                    self.maxoutcount:]
            else:
                chunk = self.pendingoutput[0]
                del self.pendingoutput[0]
            self._sendoutput(chunk, sendbreak=dobreak)

    def _sendoutput(self, output, sendbreak=False):
        self.myseq += 1
        self.myseq &= 0xf
        if self.myseq == 0:
            self.myseq = 1
        # currently we don't try to combine ack with outgoing data
        # so we use 0 for ack sequence number and accepted character
        # count
        breakbyte = 0
        if sendbreak:
            breakbyte = 0b10000
        try:
            payload = bytearray((self.myseq, 0, 0, breakbyte)) + output
        except TypeError:  # bytearray hits unicode...
            payload = bytearray((self.myseq, 0, 0, breakbyte
                                 )) + output.encode('utf8')
        self.lasttextsize = len(output)
        needskeepalive = False
        if self.lasttextsize == 0:
            needskeepalive = True
        self.awaitingack = True
        self.lastpayload = payload
        self.send_payload(payload, retry=False, needskeepalive=needskeepalive)
        retries = 5
        while retries and self.awaitingack:
            expiry = _monotonic_time() + 5.5 - retries
            while self.awaitingack and _monotonic_time() < expiry:
                self.wait_for_rsp(0.5)
            if self.awaitingack:
                self.send_payload(payload, retry=False,
                                  needskeepalive=needskeepalive)
            retries -= 1
        if not retries:
            self._print_error('Connection lost')

    def send_payload(self, payload, payload_type=1, retry=True,
                     needskeepalive=False):
        while not (self.connected or self.broken):
            session.Session.wait_for_rsp(timeout=10)
        if self.ipmi_session is None or not self.ipmi_session.logged:
            self._print_error('Session no longer connected')
            raise exc.IpmiException('Session no longer connected')
        self.ipmi_session.send_payload(payload,
                                       payload_type=payload_type,
                                       retry=retry,
                                       needskeepalive=needskeepalive)

    def _print_info(self, info):
        self._print_data({'info': info})

    def _print_error(self, error):
        self.broken = True
        if self.ipmi_session:
            self.ipmi_session.unregister_keepalive(self.keepaliveid)
            if (self.ipmi_session.sol_handler
                    and self.ipmi_session.sol_handler.__self__ is self):
                self.ipmi_session.sol_handler = None
            self.ipmi_session = None
        if type(error) == dict:
            self._print_data(error)
        else:
            self._print_data({'error': error})

    def _print_data(self, data):
        """Convey received data back to caller in the format of their choice.

        Caller may elect to provide this class filehandle(s) or else give a
        callback function that this class will use to convey data back to
        caller.
        """
        self.out_handler(data)

    def _got_sol_payload(self, payload):
        """SOL payload callback"""

        # TODO(jbjohnso) test cases to throw some likely scenarios at functions
        # for example, retry with new data, retry with no new data
        # retry with unexpected sequence number
        if type(payload) == dict:  # we received an error condition
            self.activated = False
            self._print_error(payload)
            return
        newseq = payload[0] & 0b1111
        ackseq = payload[1] & 0b1111
        ackcount = payload[2]
        nacked = payload[3] & 0b1000000
        poweredoff = payload[3] & 0b100000
        deactivated = payload[3] & 0b10000
        breakdetected = payload[3] & 0b100
        # for now, ignore overrun.  I assume partial NACK for this reason or
        # for no reason would be treated the same, new payload with partial
        # data.
        remdata = ""
        remdatalen = 0
        if newseq != 0:  # this packet at least has some data to send to us..
            if len(payload) > 4:
                remdatalen = len(payload[4:])  # store remote len before dupe
                # retry logic, we must ack *this* many even if it is
                # a retry packet with new partial data
                remdata = bytes(payload[4:])
            if newseq == self.remseq:  # it is a retry, but could have new data
                if remdatalen > self.lastsize:
                    remdata = bytes(remdata[4 + self.lastsize:])
                else:  # no new data...
                    remdata = ""
            else:  # TODO(jbjohnso) what if remote sequence number is wrong??
                self.remseq = newseq
            self.lastsize = remdatalen
            if remdata:  # Do not subject callers to empty data
                self._print_data(remdata)
            ackpayload = bytearray((0, self.remseq, remdatalen, 0))
            # Why not put pending data into the ack? because it's rare
            # and might be hard to decide what to do in the context of
            # retry situation
            try:
                self.send_payload(ackpayload, retry=False)
            except exc.IpmiException:
                # if the session is broken, then close the SOL session
                self.close()
        if self.myseq != 0 and ackseq == self.myseq:  # the bmc has something
            # to say about last xmit
            self.awaitingack = False
            if nacked and not breakdetected:  # the BMC was in some way unhappy
                if poweredoff:
                    self._print_info("Remote system is powered down")
                if deactivated:
                    self.activated = False
                    self._print_error("Remote IPMI console disconnected")
                else:  # retry all or part of packet, but in a new form
                    # also add pending output for efficiency and ease
                    newtext = self.lastpayload[4 + ackcount:]
                    with self.outputlock:
                        if (self.pendingoutput
                                and not isinstance(self.pendingoutput[0],
                                                   dict)):
                            self.pendingoutput[0] = \
                                newtext + self.pendingoutput[0]
                        else:
                            self.pendingoutput = [newtext] + self.pendingoutput
            # self._sendpendingoutput() checks len(self._sendpendingoutput)
            self._sendpendingoutput()
        elif ackseq != 0 and self.awaitingack:
            # if an ack packet came in, but did not match what we
            # expected, retry our payload now.
            # the situation that was triggered was a senseless retry
            # when data came in while we xmitted.  In theory, a BMC
            # should handle a retry correctly, but some do not, so
            # try to mitigate by avoiding overeager retries
            # occasional retry of a packet
            # sooner than timeout suggests is evidently a big deal
            self.send_payload(payload=self.lastpayload, retry=False)

    def main_loop(self):
        """Process all events until no more sessions exist.

        If a caller is a simple little utility, provide a function to
        eternally run the event loop.  More complicated usage would be expected
        to provide their own event loop behavior, though this could be used
        within the greenthread implementation of caller's choice if desired.
        """
        # wait_for_rsp promises to return a false value when no sessions are
        # alive anymore
        # TODO(jbjohnso): wait_for_rsp is not returning a true value for our
        # own session
        while (1):
            session.Session.wait_for_rsp(timeout=600)


class ServerConsole(Console):
    """IPMI SOL class.

    This object represents an SOL channel, multiplexing SOL data with
    commands issued by ipmi.command.

    :param session: IPMI session
    :param iohandler: I/O handler
    """

    def __init__(self, _session, iohandler, force=False):
        self.outputlock = threading.RLock()
        self.keepaliveid = None
        self.connected = True
        self.broken = False
        self.out_handler = iohandler
        self.remseq = 0
        self.myseq = 0
        self.lastsize = 0
        self.retriedpayload = 0
        self.pendingoutput = []
        self.awaitingack = False
        self.activated = True
        self.force_session = force
        self.ipmi_session = _session
        self.ipmi_session.sol_handler = self._got_sol_payload
        self.maxoutcount = 256
        self.poweredon = True

        session.Session.wait_for_rsp(0)

    def _got_sol_payload(self, payload):
        """SOL payload callback"""

        # TODO(jbjohnso) test cases to throw some likely scenarios at functions
        # for example, retry with new data, retry with no new data
        # retry with unexpected sequence number
        if type(payload) == dict:  # we received an error condition
            self.activated = False
            self._print_error(payload)
            return
        newseq = payload[0] & 0b1111
        ackseq = payload[1] & 0b1111
        ackcount = payload[2]
        nacked = payload[3] & 0b1000000
        breakdetected = payload[3] & 0b10000
        # for now, ignore overrun.  I assume partial NACK for this reason or
        # for no reason would be treated the same, new payload with partial
        # data.
        remdata = ""
        remdatalen = 0
        flag = 0
        if not self.poweredon:
            flag |= 0b1100000
        if not self.activated:
            flag |= 0b1010000
        if newseq != 0:  # this packet at least has some data to send to us..
            if len(payload) > 4:
                remdatalen = len(payload[4:])  # store remote len before dupe
                # retry logic, we must ack *this* many even if it is
                # a retry packet with new partial data
                remdata = bytes(payload[4:])
            if newseq == self.remseq:  # it is a retry, but could have new data
                if remdatalen > self.lastsize:
                    remdata = bytes(remdata[4 + self.lastsize:])
                else:  # no new data...
                    remdata = ""
            else:  # TODO(jbjohnso) what if remote sequence number is wrong??
                self.remseq = newseq
            self.lastsize = remdatalen
            ackpayload = bytearray((0, self.remseq, remdatalen, flag))
            # Why not put pending data into the ack? because it's rare
            # and might be hard to decide what to do in the context of
            # retry situation
            try:
                self.send_payload(ackpayload, retry=False)
            except exc.IpmiException:
                # if the session is broken, then close the SOL session
                self.close()
            if remdata:  # Do not subject callers to empty data
                self._print_data(remdata)
        if self.myseq != 0 and ackseq == self.myseq:  # the bmc has something
            # to say about last xmit
            self.awaitingack = False
            if nacked and not breakdetected:  # the BMC was in some way unhappy
                newtext = self.lastpayload[4 + ackcount:]
                with self.outputlock:
                    if (self.pendingoutput
                            and not isinstance(self.pendingoutput[0], dict)):
                        self.pendingoutput[0] = newtext + self.pendingoutput[0]
                    else:
                        self.pendingoutput = [newtext] + self.pendingoutput
            # self._sendpendingoutput() checks len(self._sendpendingoutput)
            self._sendpendingoutput()
        elif ackseq != 0 and self.awaitingack:
            # if an ack packet came in, but did not match what we
            # expected, retry our payload now.
            # the situation that was triggered was a senseless retry
            # when data came in while we xmitted.  In theory, a BMC
            # should handle a retry correctly, but some do not, so
            # try to mitigate by avoiding overeager retries
            # occasional retry of a packet
            # sooner than timeout suggests is evidently a big deal
            self.send_payload(payload=self.lastpayload)

    def send_payload(self, payload, payload_type=1, retry=True,
                     needskeepalive=False):
        while not (self.connected or self.broken):
            session.Session.wait_for_rsp(timeout=10)
        self.ipmi_session.send_payload(payload,
                                       payload_type=payload_type,
                                       retry=retry,
                                       needskeepalive=needskeepalive)

    def close(self):
        """Shut down an SOL session"""

        self.activated = False



# Copyright 2015 Lenovo
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct

categories = {}


def register_inventory_category(module):
    c = module.get_categories()
    for id in c:
        categories[id] = c[id]


class EntryField(object):
    """Store inventory field parsing options.

    Represents an inventory field and its options for the custom requests to a
    ThinkServer's BMC.

    :param name: the name of the field
    :param fmt: the format of the field (see struct module for details)
    :param include: whether to include the field in the parse output
    :param mapper: a dictionary mapping values to new values for the parse
                   output
    :param valuefunc: a function to be called to change the value in the last
                      step of the build process.
    :param presence: whether the field indicates presence. In this case, the
                     field will not be included. If the value is false, the
                     item will be discarded.
    """
    def __init__(self, name, fmt, include=True, mapper=None, valuefunc=None,
                 multivaluefunc=False, presence=False):
        self.name = name
        self.fmt = fmt
        self.include = include
        self.mapper = mapper
        self.valuefunc = valuefunc
        self.multivaluefunc = multivaluefunc
        self.presence = presence


# General parameter parsing functions
def parse_inventory_category(name, info, asrock=False, countable=True):
    """Parses every entry in an inventory category

    For example: CPU, memory, PCI, drives
    Expects the first byte to be a count of the number of entries, followed
    by a list of elements to be parsed by a dedicated parser (below).

    :param name: the name of the parameter (e.g.: "cpu")
    :param info: a list of integers with raw data read from an IPMI requests
    :param asrock: a boolean represents if RS160 with asrockrack or not
    :param countable: whether the data have an entries count field
    :returns: dict -- a list of entries in the category.
    """
    raw = info["data"][1:]

    if name == "cpu" and asrock:
        raw = info["data"]

    cur = 0
    if countable:
        count = bytearray(raw)[cur]
        cur += 1
    else:
        count = 0
    discarded = 0

    entries = []
    while cur < len(raw):
        read, parser = categories[name]["parser"](raw[cur:])
        cur = cur + read
        # Account for discarded entries (because they are not present)
        if parser is None:
            discarded += 1
            continue
        if not countable:
            # count by myself
            count += 1
            parser["index"] = count
        entries.append(parser)

    # TODO(avidal): raise specific exception to point that there's data left in
    # the buffer
    if cur != len(raw):
        raise Exception
    # TODO(avidal): raise specific exception to point that the number of
    # entries is different than the expected
    if count - discarded != len(entries) and not asrock:
        raise Exception
    return entries


def parse_inventory_category_entry(raw, fields):
    """Parses one entry in an inventory category.

    :param raw: the raw data to the entry. May contain more than one entry,
                only one entry will be read in that case.
    :param fields: an iterable of EntryField objects to be used for parsing the
                   entry.

    :returns: dict -- a tuple with the number of bytes read and a dictionary
                      representing the entry.
    """
    r = raw

    obj = {}
    bytes_read = 0
    discard = False
    for field in fields:
        value = struct.unpack_from(field.fmt, r)[0]
        read = struct.calcsize(field.fmt)
        bytes_read += read
        if bytes_read > len(raw):
            break

        r = r[read:]
        # If this entry is not actually present, just parse and then discard it
        if field.presence and not bool(value):
            discard = True
        if not field.include:
            continue

        if (field.fmt[-1] == "s"):
            value = value.rstrip(b'\x00\xff')
        if (field.mapper and value in field.mapper):
            value = field.mapper[value]
        if (field.valuefunc):
            value = field.valuefunc(value)

        if not field.multivaluefunc:
            obj[field.name] = value
        else:
            for key in value:
                obj[key] = value[key]

    if discard:
        obj = None
    return bytes_read, obj


def parse_bios_number_entry(raw):
    """Parses the Bios number given a raw data.

    :param raw: the raw data to the entry.

    :returns: dict -- structure with read current and newest versions
    """
    new_major_version = struct.unpack_from("1B", raw, 25)[0]
    new_minor_version = struct.unpack_from("1B", raw, 26)[0]
    new_aux = struct.unpack_from("I", raw, 27)[0]

    cur_major_version = struct.unpack_from("1B", raw, 31)[0]
    cur_minor_version = struct.unpack_from("1B", raw, 32)[0]
    cur_aux = struct.unpack_from("I", raw, 33)[0]

    new_image_version = "%s.%s.%s" % (
        str(new_major_version),
        str(new_minor_version),
        str(new_aux))
    cur_image_version = "%s.%s.%s" % (
        str(cur_major_version),
        str(cur_minor_version),
        str(cur_aux))

    return {
        'new_img_version': new_image_version,
        'cur_img_version': cur_image_version
    }
